"""
Functions for visualizing nodes and their links with graphviz.
"""
from graphviz import Digraph

from . import get_connected_nodes
from .nodes.base import MultiNode


def visualize_graph_module(graph, compactify=True):
    """
    Make a dot graph for a GraphModule.

    :param graph: A GraphModule
    :param compactify: see :func:`visualize_node_set`

    :returns: graphviz.Digraph
    """
    node_set = {*graph.forward_output_list, *graph.input_nodes}
    preserve = set(graph.nodes_to_compute)
    return visualize_node_set(node_set, compactify=compactify, preserve=preserve)


def visualize_connected_nodes(node_iterable, compactify=True):
    """

    Make a dot graph for a all nodes connected to a group of nodes.

    :param node_iterable: An iterable (tuple, list, etc)  of nodes.
    :param compactify: see :func:`visualize_node_set`

    :returns: graphviz.DiGraph
    """
    preserve = set(node_iterable)
    node_set = get_connected_nodes(preserve)
    return visualize_node_set(node_set, preserve=preserve, compactify=compactify)


def visualize_node_set(node_set, compactify, preserve=None):
    """

    Make a dot graph for a precise set of nodes.

    :param node_set: The set of nodes to visualize.
    :param compactify: If True, do not show the individual outputs of a `MultiNode`. If True, show all nodes.
    :param preserve:

    :returns: graphviz.Digraphs

    """

    dot = Digraph()
    starting_dot_names = get_viz_node_names(node_set)
    if compactify:
        dot_names = compactify_node_names(starting_dot_names, preserve=preserve)
    else:
        dot_names = starting_dot_names

    # Create the visible nodes
    for node in dot_names:
        if dot_names[node] == starting_dot_names[node]:
            node_attr = {"style": "rounded", "shape": "rect"}

            if isinstance(node, MultiNode):
                node_attr["style"] = "rounded,bold"

            if node.parents == ():
                node_attr["color"] = "green"
            elif preserve and node in preserve:
                node_attr["color"] = "blue"
            elif all(c not in dot_names for c in node.children):
                node_attr["color"] = "red"

            dn = dot.node(dot_names[node], **node_attr)

    # Create the links between nodes:
    for node in node_set:
        for child in node.children:
            if child in node_set:
                n1 = dot_names[node]
                n2 = dot_names[child]
                if n1 != n2:  # supports compactify option
                    dot.edge(n1, n2)

    return dot


def get_viz_node_names(node_set):
    """
    :param node_set:
    :return:
    :meta private:
    """

    base_names = set()
    nonunique_names = set()
    for node in node_set:
        name = node.name
        if name in base_names:
            nonunique_names.add(node.name)
        base_names.add(node.name)

    unique_names = {}
    for node in node_set:
        if node.name in nonunique_names:
            unique_names[node] = "{} (id={})".format(node.name, id(node))
        else:
            unique_names[node] = node.name
    return unique_names


def compactify_node_names(names, preserve):
    """
    :meta private:
    Generate a list of node names where multinode outputs
    are mapped to the base multinode name, unless that
    output is in the ``preserve`` set.

    :param names: dict mapping nodes to names.
    :param preserve: nodes to keep explicit.
    :return: new dict of nodes to names.
    """
    copy_names = names.copy()
    if preserve is None:
        preserve = set()  # Empty; nothing
    for node in names:
        if isinstance(node, MultiNode):
            for child in node.children:
                if child not in preserve:
                    copy_names[child] = names[node]

    return copy_names



import functools

import numpy as np
import warnings
import torch

from ase.calculators.calculator import compare_atoms, PropertyNotImplementedError, Calculator # Calculator is required to allow HIPNN to be used with ASE Mixing Calculators

from hippynn.graphs import find_relatives, find_unique_relative, get_subgraph, copy_subgraph, replace_node, GraphModule
from hippynn.graphs.gops import check_link_consistency

from hippynn.graphs.nodes.base.node_functions import NodeOperationError, NodeNotFound
from hippynn.graphs.nodes.base import InputNode
from hippynn.graphs.nodes.tags import Encoder, AtomIndexer, PairIndexer, Energies, Charges
from hippynn.graphs.nodes.pairs import ExternalNeighborIndexer
from hippynn.graphs.nodes.misc import StrainInducer
from hippynn.graphs.nodes.physics import CoulombEnergyNode, DipoleNode, StressForceNode
from hippynn.graphs.nodes.pairs import PairFilter

from hippynn.graphs.nodes.inputs import SpeciesNode, PositionsNode, CellNode


import ase.neighborlist

# TODO: implement neighbors using scipy.spatial.cKDTree?
# This works for orthorhombic boxes and is much faster than ASE...


def setup_ASE_graph(energy, charges=None, extra_properties=None):

    if charges is None:
        required_nodes = [energy]
    else:
        required_nodes = [energy, charges]

    if extra_properties is not None:
        extra_names = list(extra_properties.keys())
        for ename in extra_names:
            if not ename.isidentifier():
                raise ValueError("ASE properties must be a valid python identifier. (Got '{}')".format(ename))
        del ename
        required_nodes = required_nodes + list(extra_properties.values())

    why = "Generating ASE Calculator interface"
    subgraph = get_subgraph(required_nodes)

    ########################################
    # TODO: Implement Ewald, Wolf, or similar version of coulomb energy?
    # Better: figure out how to get voltages from external code and pass back as gradients for HIPNN backwards pass.
    ########################################

    ###############################################################
    # Get a new subgraph, and find the nodes we need to construct the calculator
    # Factory for seeking out nodes only in the subgraph and of a specific type
    search_fn = lambda targ, sg: lambda n: n in sg and isinstance(n, targ)

    try:
        pair_indexers = find_relatives(required_nodes, search_fn(PairIndexer, subgraph), why_desc=why)
    except NodeOperationError as ee:
        raise ValueError(
            "No Pair indexers found. Why build an ASE interface with no need for neighboring atoms?"
        ) from ee

    # The required nodes passed back are copies of the ones passed in.
    # We use assume_inputed to avoid grabbing pieces of the graph
    # that are only prerequisites for the pair indexer.
    new_required, new_subgraph = copy_subgraph(required_nodes, assume_inputed=pair_indexers, tag="ASE")
    # We now need access to the copied indexers, rather than the originals
    pair_indexers = find_relatives(new_required, search_fn(PairIndexer, new_subgraph), why_desc=why)

    species = find_unique_relative(new_required, search_fn(SpeciesNode, new_subgraph), why_desc=why)
    positions = find_unique_relative(new_required, search_fn(PositionsNode, new_subgraph), why_desc=why)

    # TODO: is .clone necessary? Or good? Or torch.as_tensor instead?
    encoder = find_unique_relative(species, search_fn(Encoder, new_subgraph), why_desc=why)
    species_set = torch.as_tensor(encoder.species_set).to(torch.int64)  # works with lists or tensors
    indexer = find_unique_relative(species, search_fn(AtomIndexer, new_subgraph), why_desc=why)
    min_radius = max(p.dist_hard_max for p in pair_indexers)
    ###############################################################

    ###############################################################
    # Set up graph to accept external pair indices and shifts

    in_shift = InputNode("(ASE)shift_vector")
    in_cell = CellNode("(ASE)cell")
    in_pair_first = InputNode("(ASE)pair_first")
    in_pair_second = InputNode("(ASE)pair_second")
    external_pairs = ExternalNeighborIndexer(
        "(ASE)EXTERNAL_NEIGHBORS",
        (positions, indexer.real_atoms, in_shift, in_cell, in_pair_first, in_pair_second),
        hard_dist_cutoff=min_radius,
    )
    new_inputs = [species, positions, in_cell, in_pair_first, in_pair_second, in_shift]

    # Construct Filters
    # Replace the existing pair indexers with the corresponding new (filtered) node
    # that accepts external pairs of atoms:
    # (This is the primary reason we needed to copy the subgraph --)
    #  we don't want to break the original computation, and `replace_node` mutates graph connectivity
    for pi in pair_indexers:
        if pi.dist_hard_max == min_radius:
            mapped_node = external_pairs
        else:
            mapped_node = PairFilter(
                "DistanceFilter-(ASE)EXTERNAL_NEIGHBORS",
                (external_pairs),
                dist_hard_max=pi.dist_hard_max, 
            )
        replace_node(pi, mapped_node, disconnect_old=True)
    ###############################################################

    ###############################################################
    # Set up gradient and, if possible, dipole properties.

    energy, *new_required = new_required

    cellscaleinducer = StrainInducer("(ASE)Strain_inducer", (positions, in_cell))
    strain = cellscaleinducer.strain
    derivatives = StressForceNode("(ASE)StressForceCalculator", (energy, strain, positions, in_cell))

    replace_node(positions, cellscaleinducer.strained_coordinates)
    replace_node(in_cell, cellscaleinducer.strained_cell)

    implemented_nodes = energy.main_output, derivatives.forces, derivatives.stress
    implemented_properties = ["potential_energy", "forces", "stress"]

    pbc_handler = PBCHandle(derivatives)

    if charges is not None:
        charges, *new_required = new_required
        dipole_moment = DipoleNode("(ASE)DIPOLE", charges)
        implemented_nodes = *implemented_nodes, charges.main_output, dipole_moment
        implemented_properties = implemented_properties + ["charges", "dipole_moment"]

    #### Add other properties here:
    if extra_properties is not None:
        implemented_nodes = *implemented_nodes, *new_required
        implemented_properties = implemented_properties + extra_names

    ###############################################################

    # Finally, assemble the graph!
    check_link_consistency((*new_inputs, *implemented_nodes))
    mod = GraphModule(new_inputs, implemented_nodes)
    mod.eval()

    return min_radius, species_set, implemented_properties, mod, pbc_handler


class PBCHandle:
    def __init__(self, *nodes):
        self.modules = [n.torch_module for n in nodes]
        self._last = None

    def set(self, value):

        if isinstance(value, np.ndarray):
            if np.all(np.equal(self._last, value)):
                return
        else:
            if value == self._last:
                return

        self._last = value

        # Convert PBC arrays into simple True/False values (until we have determined how to use mixed PBC)
        if value is not True and value is not False:
            if all(v == True for v in value):
                value = True
            elif all(v == False for v in value):
                value = False
            else:
                raise ValueError("Unrecognized PBC condition: {}. Mixed PBC not yet supported.".format(value))
        for m in self.modules:
            m.pbc = value


# decorator for generating calculation methods; this code is also run for calls to
# get_property, but with the key filled in as a closure.
def _generate_calculation_method(key):

    cant_calculate = PropertyNotImplementedError("Property not implemented:'{}'".format(key))

    def method(self, atoms, allow_calculation=True, **kwargs):

        if key not in self.implemented_properties:
            raise cant_calculate

        if not allow_calculation:
            return self.results.get(key, None)

        if self.calculation_required(atoms, [key]):
            self.calculate()

        return self.results[key]

    return method


# UNUSED, was used for HippynnCaclulator.to()
# factory for forwarding pytorch methods to the calculator
def pass_to_pytorch(fn_name):
    wraps = getattr(torch.nn.Module, fn_name)

    @functools.wraps(wraps)
    def method(self, *args, **kwargs):
        getattr(self.module, fn_name)(*args, **kwargs)

    return method


class HippynnCalculator(Calculator): # Calculator inheritance required for ASE Mixing Calculator usage
    """
    ASE calculator based on hippynn graphs. Uses ASE neighbor lists. Not suitable for domain decomposition.
    """

    def __init__(self, energy, charges=None, skin=1.0, extra_properties=None, en_unit=None, dist_unit=None):
        """
        :param energy: Node for energy
        :param charges: Node for charges (optional)
        :param skin: Skin for neighbors list
        :param extra_properties: dictionary of names to nodes for additional nodes for the calculator to compute.
        :param name: identifying string
        :param en_unit: unit factor to use for energies -- set it to the same as the energy unit used
            in training. If not given, the model is assumed to output in kcal/mol!
        :param dist_unit: unit factor to use for distances -- set it to the same as the
            distance unit used in training. If not given, defaults to Angstrom
        """

        self.min_radius, self.species_set, self.implemented_properties, self.module, self.pbc = setup_ASE_graph(
            energy, charges=charges, extra_properties=extra_properties
        )
        
        self.implemented_properties.append("energy") # Required for using mixing calculators in ASE
        
        self.atoms = None
        self._last_properties = None

        # get species set to determine length of cutoffs
        if skin < 0:
            raise ValueError("Negative skin radius not allowed.")
        self._cutoffs = self.min_radius  # The -1 is for 'blank atoms', not used in ase.
        self._skin = skin
        self.nl = None
        self.rebuild_neighbors()

        self._needs_calculation = True
        self.results = {}
        self.en_unit = en_unit if en_unit is not None else ase.units.kcal / ase.units.mol
        self.dist_unit = dist_unit if dist_unit is not None else ase.units.Angstrom
        self.device = torch.device("cpu")
        self.dtype = torch.get_default_dtype()
        if not hasattr(self, "name"):  # Older versions of ase do not set the name
            self.name = "Hippynn calculator"
        self.parameters = {} #Hack to work with ASE trajectory printing

    make = _generate_calculation_method
    get_potential_energy = make("potential_energy")
    get_energy = get_potential_energy
    get_potential_energies = make("potential_energies")
    get_energies = get_potential_energies
    get_free_energy = make("free_energy")
    get_forces = make("forces")
    get_charges = make("charges")
    get_dipole_moment = make("dipole_moment")
    get_dipole = get_dipole_moment
    get_stress = make("stress")
    get_stresses = make("stresses")
    get_magmom = make("magmom")
    get_magmoms = make("magmoms")
    del make

    def rebuild_neighbors(self):
        self.nl = ase.neighborlist.NeighborList(
            self._cutoffs+self._skin, #ASE neighbor list implementation is only safe up to cutoffs-skin
            skin=self._skin,
            sorted=True,
            self_interaction=False,
            bothways=True,
            primitive=ase.neighborlist.NewPrimitiveNeighborList,
        )

    # Dear ASE: This is not Java....
    def set_atoms(self, atoms):
        self.atoms = atoms.copy()
        self._needs_calculation = True

    def get_property(self, name, atoms, allow_calculation=True):
        # `or` defaults to last atoms received
        try:
            return getattr(self, "get_{}".format(name))(atoms or self.atoms, allow_calculation=allow_calculation)
        except AttributeError:
            raise PropertyNotImplementedError("Property not implemented:'{}'".format(name))

    def to(self, *args, **kwargs):
        self.module.to(*args, **kwargs)

        # The below section is complicated because the pytorch `to` method does not have a strict signature.
        allargs = *args, *kwargs.values()
        for a in allargs:
            # Attempt conversion of strings to dtype or device.
            try:
                a = torch.device(a)
            except (RuntimeError, TypeError):
                pass
            try:
                a = torch.dtype(a)
            except (RuntimeError, TypeError):
                pass

            if isinstance(a, torch.dtype):
                self.dtype = a
            if isinstance(a, torch.device):
                self.device = a
        return self

    def calculate(self, atoms=None, properties=None, system_changes=True):
        """
        Accepts 'properties' and 'system changes' but ignores them,
        purely for compatibility with ASE base calculator
        """
        self.atoms = self.atoms if atoms is None else atoms.copy()

        self.pbc.set(self.atoms.pbc)

        # Ase neighbor list raises ValueError if passed a new system with a different number of atoms.
        # If something like this happens, we just scrap that neighbors list and make a fresh one.
        try:
            self.nl.update(self.atoms)
        except ValueError:
            self.rebuild_neighbors()
            self.nl.update(self.atoms)

        # Get variables from atoms. Unsqueeze is to add batch axis.
        positions = torch.as_tensor(self.atoms.positions).unsqueeze(0)
        # Convert from ASE distance (angstrom) to whatever the network uses.
        positions = positions / self.dist_unit
        species = torch.as_tensor(self.atoms.numbers,dtype=torch.long).unsqueeze(0)
        cell = torch.as_tensor(self.atoms.cell.array)  # ExternalNieghbors doesn't take batch index
        # Get pair first and second from neighbors list

        pair_first = torch.as_tensor(self.nl.nl.pair_first,dtype=torch.long)
        pair_second = torch.as_tensor(self.nl.nl.pair_second,dtype=torch.long)
        pair_shiftvecs = torch.as_tensor(self.nl.nl.offset_vec,dtype=torch.long)

        # This order must be synchronized with function setup_ase_graph above
        inputs = species, positions, cell, pair_first, pair_second, pair_shiftvecs
        # Move to device, and convert to the type of float the model is using
        inputs = [
            inp.to(device=self.device, dtype=(self.dtype if torch.is_floating_point(inp) else None)) for inp in inputs
        ]

        # Run it all through pytorch!
        results = self.module(*inputs)

        self.results = {k: r.detach().cpu().numpy() for k, r in zip(self.implemented_properties, results)}

        # Convert units
        self.results["potential_energy"] = self.results["potential_energy"][0, 0] * self.en_unit
        self.results["energy"] = self.results["potential_energy"] # Required for using ASE mixing calculators, which assume potential energy is under the property "energy"
        self.results["forces"] = self.results["forces"][0] * (self.en_unit / self.dist_unit)
        # slightly opaque way to handle if pbc is 3-tuple or boolean.
        # Note: PBC handler forbids mixed BCs, so this check is enough.
        if all(self.atoms.pbc) if len(self.atoms.pbc) else self.atoms.pbc:
            stress_factor = self.en_unit / (self.dist_unit) ** 3
        else:
            stress_factor = self.en_unit
        self.results["stress"] = self.results["stress"][0] * stress_factor

        self._needs_calculation = False

    def calculation_required(self, atoms, properties=None, tol=1e-15):
        """
        Returns true if:
        1. A property in the list `properties` is not supported.
        2. Atoms are different from stored atoms. This proceeds by short-circuit:
        * if this check is passed, compare using ase `compare_atoms` and tolerance `tol`.
        * slight difference from ASE implementation: if `properties` is a string then it is wrapped in a list.

        """
        if isinstance(properties, str):
            properties = [properties]

        if any(prop not in self.implemented_properties for prop in properties):
            return True

        if bool(compare_atoms(self.atoms, atoms, tol=tol)):
            self._needs_calculation = True
            self.atoms = atoms.copy()

        return self._needs_calculation


def calculator_from_model(model, **kwargs):
    """
    Attempt to find the energy and charge nodes automatically.

    :param model: :class:`GraphModule` for the model.
    :param kwargs: passed to :class:`HippynnCalculator`

    .. Note::
       If your model has an energy node, but that is not the full energy for simulation, don't use this function.
       Similarly for charge.
    """
    possible_energy = find_relatives(model.nodes_to_compute, Energies)
    if len(possible_energy) != 1:
        raise ValueError("More than one energy node present, cannot auto-generate ASE interface")
    energy = possible_energy.pop()

    try:
        possible_charge = find_relatives(model.nodes_to_compute, Charges)
    except NodeNotFound as nnf:
        warnings.warn("No charge node found: ASE interface will not calculate charges or dipole moments.")
        charges = None
    else:
        if len(possible_charge) > 1:
            warnings.warn("More than one charge node present, charges left out of ASE interface")
            charges = None
        else:
            charges = possible_charge.pop()

    return HippynnCalculator(energy=energy, charges=charges, **kwargs)



"""
Layers for physical operations
"""
import warnings

import torch
from torch import Tensor

from . import indexers, pairs


class Gradient(torch.nn.Module):
    def __init__(self, sign):
        super().__init__()
        assert sign in (-1, 1), "Sign of gradient must be +1 (gradient) or -1 (force)"
        self.sign = sign

    def forward(self, molecular_energies, positions):
        return self.sign * torch.autograd.grad(molecular_energies.sum(), positions, create_graph=True)[0]
        
class MultiGradient(torch.nn.Module):
    def __init__(self, signs):
        super().__init__()
        if isinstance(signs, int):
            signs = (signs,)
        for sign in signs:
            assert sign in (-1,1), "Sign of gradient must be -1 or +1"
        self.signs = signs

    def forward(self, molecular_energies: Tensor, *generalized_coordinates: Tensor):
        if isinstance(generalized_coordinates, Tensor):
            generalized_coordinates = (generalized_coordinates,)
        assert len(generalized_coordinates) == len(self.signs), f"Number of items to take derivative w.r.t ({len(generalized_coordinates)}) must match number of provided signs ({len(self.signs)})."
        grads = torch.autograd.grad(molecular_energies.sum(), generalized_coordinates, create_graph=True)
        return tuple((sign * grad for sign, grad in zip(self.signs, grads)))

class StressForce(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pbc = False

    def forward(self, energy, strain, coordinates, cell):
        total_energy = energy.sum()
        straingrad, grad = torch.autograd.grad(total_energy, [strain, coordinates], create_graph=self.training)
        if self.pbc:
            stress = straingrad / torch.det(cell)
        else:
            stress = straingrad
        return -grad, stress


class Dipole(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.summer = indexers.MolSummer()

    def forward(self, charges: Tensor, positions: Tensor, mol_index: Tensor, n_molecules: int):
        if charges.shape[1] > 1:
            # charges contain multiple targets, so set up broadcasting
            charges = charges.unsqueeze(2)
            positions = positions.unsqueeze(1)

        # shape is (n_atoms, 3, n_targets) in multi-target mode
        # shape is (n_atoms, 3) in single target mode
        dipole_elements = charges * positions
        dipoles = self.summer(dipole_elements, mol_index, n_molecules)
        return dipoles


class Quadrupole(torch.nn.Module):
    """Computes quadrupoles as a flattened (n_molecules,9) array.
    NOTE: Uses normalization sum_a q_a (r_a,i*r_a,j - 1/3 delta_ij r_a^2)"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.summer = indexers.MolSummer()

    def forward(self, charges, positions, mol_index, n_molecules):
        # positions shape: (atoms, xyz)
        # charge shape: (atoms,1)
        ri_rj = positions.unsqueeze(1) * positions.unsqueeze(2)
        ri_rj_flat = ri_rj.reshape(-1, 9)  # Flatten to component
        rsq = (positions**2).sum(dim=1).unsqueeze(1)  # unsqueeze over component index
        delta_ij = torch.eye(3, device=rsq.device).flatten().unsqueeze(0)  # unsqueeze over atom index
        quad_elements = charges * (ri_rj_flat - (1 / 3) * (rsq * delta_ij))
        quadrupoles = self.summer(quad_elements, mol_index, n_molecules)
        return quadrupoles


class CoulombEnergy(torch.nn.Module):
    """ Computes the Coulomb Energy of the molecule/configuration. 
    
    Coulomb energies is defined for pairs of atoms. Here, we adopt the 
    convention that the Coulomby energy for a pair of atoms is evenly
    partitioned to both atoms as the 'per-atom energies'. Therefore, the 
    atom energies sum to the molecular energy; similar to the HEnergy. 
    """
    def __init__(self, energy_conversion_factor):
        super().__init__()
        self.register_buffer("energy_conversion_factor", torch.tensor(energy_conversion_factor))
        self.summer = indexers.MolSummer()

    def forward(self, charges, pair_dist, pair_first, pair_second, mol_index, n_molecules):
        voltage_pairs = self.energy_conversion_factor * (charges[pair_second] / pair_dist.unsqueeze(1))
        n_atoms, _ = charges.shape
        voltage_atom = torch.zeros((n_atoms, 1), device=charges.device, dtype=charges.dtype)
        voltage_atom.index_add_(0, pair_first, voltage_pairs)
        coulomb_atoms = 0.5*voltage_atom * charges
        coulomb_molecule = self.summer(coulomb_atoms, mol_index, n_molecules)
        return coulomb_molecule, coulomb_atoms, voltage_atom


class ScreenedCoulombEnergy(CoulombEnergy):
    """ Computes the Coulomb Energy of the molecule/configuration. 
    
    The convention for the atom energies is the same as CoulombEnergy
    and the HEnergy. 
    """
    
    def __init__(self, energy_conversion_factor, screening, radius=None):
        super().__init__(energy_conversion_factor)
        if screening is None:
            raise ValueError("Screened Coulomb requires specification of a screening type.")
        if radius is None:
            raise ValueError("Screened Coulomb requires specification of a radius")

        if isinstance(screening, type):
            screening = screening()

        self.radius = radius

        self.screening = screening
        self.bond_summer = pairs.MolPairSummer()

    def forward(self, charges, pair_dist, pair_first, pair_second, mol_index, n_molecules):
        screening = self.screening(pair_dist, self.radius).unsqueeze(1)
        screening = torch.where((pair_dist < self.radius).unsqueeze(1), screening, torch.zeros_like(screening))

        # Voltage pairs for per-atom energy
        voltage_pairs = self.energy_conversion_factor * (charges[pair_second] / pair_dist.unsqueeze(1)) 
        voltage_pairs = voltage_pairs * screening 
        n_atoms, _ = charges.shape
        voltage_atom = torch.zeros((n_atoms, 1), device=charges.device, dtype=charges.dtype)
        voltage_atom.index_add_(0, pair_first, voltage_pairs) 
        coulomb_atoms = 0.5 * voltage_atom * charges
        coulomb_molecule = self.summer(coulomb_atoms, mol_index, n_molecules)

        return coulomb_molecule, coulomb_atoms, voltage_atom


class CombineScreenings(torch.nn.Module):
    """ Returns products of different screenings for Screened Coulomb Interactions.
    """
    def __init__(self, screening_list):
        super().__init__()
        self.SL = torch.nn.ModuleList(screening_list)

    def forward(self, pair_dist, radius):
        """ Product of different screenings applied to pair_dist upto radius.

        :param pair_dist: torch.tensor, dtype=float64: 'Neighborlist' distances for coulomb energies.
        :param radius: Maximum radius that Screened-Coulomb is evaluated upto.
        :return screening: Weights for screening for all pair_dist.
        """
        screening = None

        for s in self.SL:
            if screening is None:
                screening = s(pair_dist=pair_dist, radius=radius)
            else:
                screening = screening * s(pair_dist=pair_dist, radius=radius)

        return screening


class AlphaScreening(torch.nn.Module):
    def __init__(self, alpha):
        super().__init__()
        self.alpha = alpha


# Note: This is somewhat incomplete as it does not include a k-space contribution -- more is needed
class EwaldRealSpaceScreening(AlphaScreening):
    def __init__(self, alpha):
        warnings.warn("Ewald implementation incomplete, does not include k-space contributions.")
        super().__init__(alpha)

    def forward(self, pair_dist, radius):
        q = pair_dist / radius
        eta = self.alpha * radius
        return torch.erfc(eta * q)


# Note: typically
class WolfScreening(AlphaScreening):
    def __init__(self, alpha):
        warnings.warn("Wolf implemnetation uses exact derivative of the potential.")
        super().__init__(alpha)

    def forward(self, pair_dist, radius):
        q = pair_dist / radius
        eta = self.alpha * radius
        return torch.erfc(eta * q) - q * torch.erfc(eta)


class LocalDampingCosine(AlphaScreening):
    """ Local damping using complement of the hipnn cutoff function. ('glue-on' method)
    g =     1 if pair_dist > R_cutoff
            1 - [cos(\pi/2 * dist * R_cutoff)]^2  otherwise
    """
    def __init__(self, alpha): 
        """ 
        :param alpha: R_cutoff for glue-on function to ensure 
            smooth crossover from hipnn energy to long-range coulomb energy.  
        """
        super().__init__(alpha) 


    def forward(self, pair_dist, radius):
        """
        :param pair_dist: torch.tensor, dtype=float64: 'Neighborlist' distances for coulomb energies.
        :param radius: Maximum radius that Screened-Coulomb is evaluated upto. 
        :return screening: Weights for screening for each pair.
        """
        pi = torch.tensor([3.141592653589793238], device=pair_dist.device)        
        screening = torch.subtract(torch.tensor([1.0], device=pair_dist.device), torch.square(torch.cos(0.5*pi*pair_dist/self.alpha)))
    
        # pair_dist greater than cut-off; no local-damping. 
        screening = torch.where((pair_dist<self.alpha), screening, torch.ones_like(screening))
        
        return screening


class QScreening(torch.nn.Module):
    def __init__(self, p_value):
        super().__init__()
        self.p_value = p_value

    @property
    def p_value(self):
        return self._p_value

    @p_value.setter
    def p_value(self, value):
        value = int(value)
        self._p_value = value
        powers = torch.arange(1, value + 1, dtype=torch.long).unsqueeze(0)
        self.register_buffer("powers", powers)

    def forward(self, pair_dist, radius):
        q = pair_dist / radius
        q_factors = 1 - torch.pow(q.unsqueeze(1), self.powers)
        product = q_factors.prod(dim=1)
        return product


class PerAtom(torch.nn.Module):
    def forward(self, features, species):
        n_atoms = (species != 0).type(features.dtype).sum(dim=1)
        return features / n_atoms.unsqueeze(1)


class VecMag(torch.nn.Module):
    def forward(self, vector_feature):
        return torch.norm(vector_feature, dim=1)


class CombineEnergy(torch.nn.Module):
    """
    Combines the energies (molecular and atom energies) from two different 
    nodes, e.g. HEnergy, Coulomb, or ScreenedCoulomb Energy Nodes. 
    """
    def __init__(self):
        super().__init__()
        self.summer = indexers.MolSummer()

    def forward(self, atom_energy_1, atom_energy_2, mol_index, n_molecules):
        """
        :param: atom_energy_1 per-atom energy from first node. 
        :param: atom_energy_2 per atom energy from second node. 
        :param: mol_index the molecular index for atoms in the batch
        :param: total number of molecules in the batch
        :return: Total Energy
        """
        total_atom_energy = atom_energy_1 + atom_energy_2
        mol_energy = self.summer(total_atom_energy, mol_index, n_molecules)
        
        return mol_energy, total_atom_energy



"""
Dataset stored as ase database file (ideally either .json or .db, or in .xyz, .extxyz formats)

See: https://databases.fysik.dtu.dk/ase/ase/db/db.html for documentation
on typical columns present in ase database

Typically used column names in ase db/xyz format:

positions : x,y,z cartesian positions (Angstrom)
forces : x,y,z carteisian forces (ev/Angstrom)
energy : energy (eV)
energy_per_atom : energy per atom (eV/atom)
cell : x,y,z of cell (3x3) (Angstrom)
charges : atom-specific charges
stress: (6,) atomic stresses
initial_charges : atom-specific initial charges 
inital_magmoms :atom-specific initial magnetic moments
numbers : atom Zs (integer)
pbc : periodic boundary conditions (bool)
ctime : computer time (float)
mtime : time (float)
dipole : molecular dipole (3) vector
"""
import os

import numpy as np
from ase.io import read

from ...tools import np_of_torchdefaultdtype
from ...databases.database import Database
from ...databases.restarter import Restartable
from typing import Union
from typing import List


class AseDatabase(Database, Restartable):
    """
    Database stored as ase database file(s) in a directory.

    :param directory: directory path where the ase database(s) is stored
    :param name: name or list of names for files for ase databases.

    This function loads an ase database(s) ({name}.json/.db) OR ({{name}.extxyz,.xyz})
    variable db_name including all inputs and targets.

    filenames should end with .json, .db, .extxyz, and .xyz, etc; Anything parsable by ase.io.load


    Other arguments: See ``Database``.

    See: https://databases.fysik.dtu.dk/ase/ase/db/db.html for documentation
        on typical columns present in ase database
    """

    def __init__(self, directory: str, name: Union[str, List[str]], inputs, targets, *args, quiet=False, allow_unfound=False, **kwargs):

        arr_dict = self.load_arrays(directory, name, inputs, targets, quiet=quiet, allow_unfound=allow_unfound)
        super().__init__(arr_dict, inputs, targets, *args, **kwargs, quiet=quiet, allow_unfound=allow_unfound)

        self.restarter = self.make_restarter(
            directory,
            name,
            inputs,
            targets,
            *args,
            **kwargs,
            quiet=quiet,
            allow_unfound=allow_unfound,
        )

    def load_arrays(self, directory, filename, inputs, targets, quiet=False, allow_unfound=False):
        """load arrays load ase database into hippynn database arrays

        Parameters
        ----------
        filename : str
            filename or path of database to convert
        prefix : str, optional
            prefix for output numpy arrays, by default None
        return_data : bool, optional
            whether or not to return the data or write to files, by default False
        """
        var_list = inputs + targets
        try:
            if isinstance(filename, str):
                db = read(directory + filename, index=":")
            elif isinstance(filename, (list, np.ndarray)):
                db = []
                for name in filename:
                    temp_db = read(directory + name, index=":")
                    db += temp_db
        except FileNotFoundError as fee:
            raise FileNotFoundError(
                "ERROR: Couldn't find {} ase xyz database."
                'A solution is to explicitly specify "path" in database_params '.format(directory + filename)
            ) from fee
        if not quiet:
            print("ASE Database found")
        record_list = []
        max_n_atom = 0
        max_atoms_record = None
        for row in db:
            result_dict = row.__dict__
            for k, v in result_dict["arrays"].items():
                result_dict[k] = v
            for k, v in result_dict["info"].items():
                result_dict[k] = v
            del result_dict["arrays"]
            del result_dict["info"]
            result_dict["cell"] = result_dict["_cellobj"][:]
            del result_dict["_cellobj"]
            if result_dict.get("_calc", None) is not None:
                calc_dict = result_dict.get("_calc").__dict__
                # Overwrite atoms-stored objects with calculator objects.
                # Helpful for .json/.db files where energy is not stored under 'info' section
                for k, v in calc_dict["results"].items():
                    result_dict[k] = v
                del result_dict["_calc"]
            if not allow_unfound:
                if "energy_per_atom" in var_list:
                    var_list += ["energy"]
                result_dict = {k: v for k, v in result_dict.items() if (k in var_list)}
            record_list.append(result_dict)
            if len(result_dict["positions"]) > max_n_atom:
                max_n_atom = len(result_dict["positions"])
                max_atoms_record = result_dict
        n_record = len(record_list)
        array_dict = dict()
        delete_cols = []
        for key, val in max_atoms_record.items():
            if isinstance(val, np.ndarray):
                array_dict[key] = np.zeros([n_record] + [x for x in val.shape])
            else:
                if isinstance(val, (float, int)):
                    array_dict[key] = np.zeros([n_record])
                    if (key == "energy") and (("energy_per_atom" in var_list) or (allow_unfound)):
                        array_dict["energy_per_atom"] = np.zeros([n_record])
                else:
                    array_dict[key] = None  # Do Not Save
                    delete_cols.append(key)
        # Save the record uid from the ase db object names.
        record_list.sort(key=lambda rec: len(rec["numbers"]))
        for i, record in enumerate(record_list):
            natom = len(record["numbers"])
            for k, v in record.items():
                if isinstance(v, np.ndarray):
                    if array_dict.get(k,None) is not None:
                        shape = array_dict[k].shape
                    else:
                        shape=[0]
                    # Note this assumes the maximum number of atoms greater than the length of property of interest
                    # E.g. 3 for dipole (make sure your training set has something with more than 3 atoms)
                    # Or 6 for stress tensor (make sure your training set has something with more than 6 atoms)
                    if (len(shape) == 2) and (shape[1] < max_n_atom):  # 1D array, e.g. dipole, stress tensor
                        array_dict[k][i, :] = v
                    elif (len(shape) == 2) and (shape[1] == max_n_atom):  # 1D array including maximum number of atoms: e.g. charges
                        array_dict[k][i, :natom] = v
                    elif (len(shape) == 3) and (shape[1] < max_n_atom):  # 2D array, e.g. cells
                        array_dict[k][i, :, :] = v
                    elif (len(shape) == 3) and (shape[1] == max_n_atom):  # 2D array, e.g. positions, forces
                        array_dict[k][i, :natom, :] = v
                    elif (len(shape) == 1):
                        print('Skipping {}'.format(k))
                    else:
                        raise ValueError("Shape of Numpy array for key: {} unknown.".format(k))
                elif isinstance(array_dict.get(k,None), np.ndarray):  # Energy, float or integers only
                    array_dict[k][i] = v
                    if (k == "energy") and (("energy_per_atom" in var_list) or (allow_unfound)):  # Add in per-atom-energy
                        array_dict["energy_per_atom"][i] = v / natom
                else:  # Everything else either list of strings or something else.
                    pass

        # Make sure Zs are integers.
        array_dict["numbers"] = array_dict["numbers"].astype("int")

        arr_dict = dict()

        # Put float64 data in pytorch default dtype
        floatX = np_of_torchdefaultdtype()
        for k, v in array_dict.items():
            if k not in delete_cols:
                if v.dtype == "float64":
                    arr_dict[k] = v.astype(floatX)
                else:
                    arr_dict[k] = v

        if not quiet:
            print("Data types:")
            print({k: v.dtype for k, v in arr_dict.items()})

        return arr_dict



"""

Functions for training.

"""
from . import controllers
from . import evaluator
from . import metric_tracker
from . import assembly
from . import routines

from .assembly import assemble_for_training
from .routines import setup_and_train, setup_training, train_model, test_model, SetupParams

__all__ = ["assemble_for_training", "setup_and_train", "setup_training", "train_model", "test_model", "SetupParams"]



"""
Modules for analyzing pair-valued data
"""
import torch


class RDFBins(torch.nn.Module):
    def __init__(self, bins, species_set):
        super().__init__()
        if bins is None:
            raise TypeError("Bins must not be None!")
        bins = torch.as_tensor(bins)
        species_set = torch.as_tensor(species_set)
        self.register_buffer("bins", bins.to(torch.float))
        self.register_buffer("species_set", species_set.to(torch.int))

    def bin_info(self):
        # Note: widths don't make perfect sense for non-evenly-spaced bins.
        centers = (self.bins[1:] + self.bins[:-1]) / 2
        widths = self.bins[1:] - self.bins[:-1]
        return centers, widths

    def forward(self, pair_dist, pair_first, pair_second, one_hot, n_molecules):
        n_species = one_hot.shape[-1]
        n_bins = self.bins.shape[0] - 1

        rdf = torch.zeros((n_species, n_species, n_bins), dtype=pair_dist.dtype, device=pair_dist.device)
        for i in range(n_species):
            for j in range(n_species):
                mask = one_hot[:, i][pair_first] & one_hot[:, j][pair_second]
                maskpairs = pair_dist[mask]
                less = maskpairs.unsqueeze(-1) < self.bins.unsqueeze(0)
                less_counts = less.sum(dim=0)
                rdf[i, j] = less_counts[..., 1:] - less_counts[..., :-1]
        return (rdf / n_molecules).unsqueeze(0)


def min_dist_info(rij_list, j_list, mol_index, atom_index, inv_real_atoms, n_atoms_max, n_molecules):
    n_atoms = rij_list.shape[0]
    dev = rij_list.device

    if rij_list.shape[1] == 0:
        # empty neighbors list
        min_dist_mol = torch.zeros(n_molecules, dtype=rij_list.dtype, device=dev)
        min_dist_atom = torch.zeros(n_atoms, dtype=rij_list.dtype, device=dev)
        min_dist_mol_atom_locs = torch.zeros(n_molecules, dtype=torch.int64, device=dev)
        min_dist_atomneigh = torch.zeros((n_atoms), dtype=torch.int64, device=dev)
        return min_dist_mol, min_dist_mol_atom_locs, min_dist_atom, min_dist_atomneigh

    rmag_list = rij_list.norm(dim=2)
    maxr = rmag_list.max()

    rmaglist_new = rmag_list.clone()
    rmaglist_new[rmaglist_new == 0] = maxr
    min_dist_atom, where_min_dist_atom = rmaglist_new.min(dim=1)

    ara = torch.arange(n_atoms, dtype=where_min_dist_atom.dtype, device=dev)
    min_dist_atomneigh = j_list[ara, where_min_dist_atom]

    min_dist_molatom = torch.full((n_molecules, n_atoms_max), maxr, device=rmag_list.device, dtype=rmag_list.dtype)
    min_dist_molatom[mol_index, atom_index] = min_dist_atom
    min_dist_mol, where_min_dist_mol = min_dist_molatom.min(dim=1)

    atom1 = where_min_dist_mol

    atom1_batchloc = torch.arange(n_molecules, device=dev, dtype=torch.int64) * n_atoms_max + atom1
    atom1_atomloc = inv_real_atoms[atom1_batchloc]
    atom2 = atom_index[min_dist_atomneigh[atom1_atomloc]]

    min_dist_mol_atom_locs = torch.stack([atom1, atom2], dim=1)

    return min_dist_mol, min_dist_mol_atom_locs, min_dist_atom, min_dist_atomneigh


class MinDistModule(torch.nn.Module):
    def forward(self, rmag_list, j_list, mol_index, atom_index, inv_real_atoms, n_atoms_max, n_molecules):
        return min_dist_info(rmag_list, j_list, mol_index, atom_index, inv_real_atoms, n_atoms_max, n_molecules)



"""
Implementation of HIPNN.
"""

import numpy as np
import torch

from typing import Union, List

from ..layers.hiplayers import (
    GaussianSensitivityModule,
    InverseSensitivityModule,
    InteractLayer,
    InteractLayerVec,
    InteractLayerQuad,
)
from ..layers.transform import ResNetWrapper


# computes E0 for the energy layer.
def compute_hipnn_e0(encoder, Z_Data, en_data, peratom=False, fit_dtype=torch.float64):
    """

    :param encoder: encoder of species to features (one-hot representation, probably)
    :param Z_Data: species data
    :param en_data: energy data
    :param peratom: whether energy is per-atom or total
    :return: energy per species as shape (n_features_encoded, 1)
    """

    original_dtype = en_data.dtype
    en_data = en_data.to(fit_dtype)
    x, nonblank = encoder(Z_Data)

    sums = x.sum(dim=1).to(en_data.dtype)
    if peratom:
        atom_counts = nonblank.sum(dim=1, keepdims=True).to(en_data.dtype)
        Z_matrix = sums / atom_counts
    else:
        Z_matrix = sums

    # Want to solve Z e = E  where Z is a composition matrix
    # Z^T Z e  = Z^T E is lower dimensional
    # e = (Z^T Z)^-1 Z^T E
    # shape n_species, n_species
    ZTZ = Z_matrix.T @ Z_matrix
    ZTZ_inv = torch.pinverse(ZTZ)
    # shape n_species, n_examples
    ZTZ_inv_Z = ZTZ_inv @ Z_matrix.T

    # shape (n_species, n_targets) (n_targets may be omitted)
    e_per_species = ZTZ_inv_Z @ en_data

    if e_per_species.ndim != 1:
        # if n_targets is included
        e_per_species = e_per_species.T

    e_per_species = e_per_species.to(original_dtype)
    return e_per_species


class Hipnn(torch.nn.Module):
    """
    Hipnn Main Module
    """

    _interaction_class = InteractLayer

    def __init__(
        self,
        n_features,
        n_sensitivities,
        dist_soft_min,
        dist_soft_max,
        dist_hard_max,
        n_atom_layers,
        n_interaction_layers=None,
        possible_species=None,
        n_input_features=None,
        sensitivity_type="inverse",
        resnet=True,
        activation=torch.nn.Softplus,
    ):
        """

        :param n_features: width of each layer
        :param n_sensitivities: number of sensitivity functions
        :param dist_soft_min: midpoint of first sensitivity function
        :param dist_soft_max: midpoint of last sensitivity function
        :param dist_hard_max: cutoff for cutoff function.
        :param n_interaction_layers: number of interaction blocks
        :param n_atom_layers: number of atom layers per interaction block.
        :param possible_species: list of species values in database, including 0 for none.
            For example, HCNO would be [0 1 6 7 8]
        :param n_input_features: number of input features to the model
        :param sensitivity_type: str or callable, type of sensitivity, default of
           'inverse' is what is in hip-nn original paper.
        :param resnet: bool or int, if int, size of internal resnet width
        :param activation: activation function or subclass of nn.module.

        Note: only one of possible_species or n_input_features is needed. If both are supplied,
        they must be consistent with each other.
        """

        super().__init__()

        if possible_species is not None:
            possible_species = torch.as_tensor(possible_species)
            if n_input_features is not None:
                if len(possible_species) - 1 != n_input_features:
                    raise ValueError("Species and input features are not consistent with each other.")
            else:
                n_input_features = len(possible_species) - 1
        else:
            if n_input_features is None:
                raise ValueError("Either n_input_features or possible_species must be set!")

        self.species_set = possible_species
        self.n_layers_per_block = n_atom_layers
        self.nf = n_features
        self.nf_in = n_input_features

        if isinstance(self.nf, int):
            if n_interaction_layers is None:
                raise ValueError("Must provide n_interaction layers if n_features is a single integer.")
            self.feature_sizes = (self.nf_in, *(self.nf for _ in range(n_interaction_layers)))
        else:
            if n_interaction_layers is not None:
                if len(self.nf) != n_interaction_layers:
                    raise ValueError("Number of interaction layers conflicts with feature sizes")
            else:
                n_interaction_layers = len(self.nf)

            self.feature_sizes = (self.nf_in, *self.nf)

        self.ni = n_interaction_layers
        self.resnet: Union[int, List[int]] = resnet

        if self.resnet not in (True, False):
            # resnet argument specifies a different size for the internal layers
            if isinstance(self.resnet, int):
                self.nf_middle = [self.resnet for _ in range(self.ni)]
            else:
                if len(self.resnet) != self.ni:
                    raise ValueError("Number of interaction layers conflicts with resnet size")
                self.nf_middle = self.resnet

        else:
            self.nf_middle = self.feature_sizes[1:]

        self.n_sensitivities = n_sensitivities
        self.dist_soft_max = dist_soft_max
        self.dist_soft_min = dist_soft_min
        self.dist_hard_max = dist_hard_max

        if not isinstance(activation, torch.nn.Module):
            activation = activation()
        self.activation = activation

        # Containers for main layers
        self.blocks = torch.nn.ModuleList()
        self.atomlayers = torch.nn.ModuleList()
        self.interactlayers = torch.nn.ModuleList()

        # determine sensitivity layer
        if sensitivity_type == "inverse":
            sensitivity_type = InverseSensitivityModule
        elif sensitivity_type == "linear":
            sensitivity_type = GaussianSensitivityModule
        elif callable(sensitivity_type):
            pass
        else:
            raise TypeError("Invalid sensitivity type:", sensitivity_type)

        # Finally, build the network!
        for in_size, out_size, middle_size in zip(self.feature_sizes[:-1], self.feature_sizes[1:], self.nf_middle):
            this_block = torch.nn.ModuleList()

            # Add interaction layer
            lay = self._interaction_class(
                in_size, middle_size, n_sensitivities, dist_soft_min, dist_soft_max, dist_hard_max, sensitivity_type
            )
            if self.resnet:
                lay = ResNetWrapper(lay, in_size, middle_size, out_size, self.activation)
            this_block.append(lay)

            for j in range(n_atom_layers):
                # Add subsequent atom layers
                lay = torch.nn.Linear(out_size, middle_size)
                torch.nn.init.xavier_normal_(lay.weight.data)
                if self.resnet:
                    lay = ResNetWrapper(lay, out_size, middle_size, out_size, self.activation)
                this_block.append(lay)
            self.blocks.append(this_block)

    @property
    def interaction_layers(self):
        return [block[0] for block in self.blocks]

    @property
    def sensitivity_layers(self):
        if self.resnet:
            return [il.base_layer.sensitivity for il in self.interaction_layers]
        else:
            return [il.sensitivity for il in self.interaction_layers]

    def regularization_params(self):
        params = []
        for block in self.blocks:
            for lay in block:
                if hasattr(lay, "regularization_params"):
                    params.extend(lay.regularization_params())
                else:
                    params.append(lay.weight)
        params = list(set(params))
        return params

    def forward(self, features, pair_first, pair_second, pair_dist):
        features = features.to(pair_dist.dtype)  # Convert one-hot features to floating point features.

        if pair_dist.ndim == 2:
            pair_dist = pair_dist.squeeze(dim=1)

        output_features = [features]

        for block in self.blocks:
            int_layer = block[0]
            atom_layers = block[1:]

            features = int_layer(features, pair_first, pair_second, pair_dist)
            if not self.resnet:
                features = self.activation(features)
            for lay in atom_layers:
                features = lay(features)
                if not self.resnet:
                    features = self.activation(features)
            output_features.append(features)

        return output_features


class HipnnVec(Hipnn):
    """
    HIP-NN-TS with l=1
    """

    _interaction_class = InteractLayerVec

    def forward(self, features, pair_first, pair_second, pair_dist, pair_coord):
        features = features.to(pair_dist.dtype)  # Convert one-hot features to floating point features.

        if pair_dist.ndim == 2:
            pair_dist = pair_dist.squeeze(dim=1)

        if pair_coord.ndim == 3:
            pair_coord = pair_coord.squeeze(dim=2)

        output_features = [features]

        for block in self.blocks:
            int_layer = block[0]
            atom_layers = block[1:]

            features = int_layer(features, pair_first, pair_second, pair_dist, pair_coord)
            if not self.resnet:
                features = self.activation(features)
            for lay in atom_layers:
                features = lay(features)
                if not self.resnet:
                    features = self.activation(features)
            output_features.append(features)

        return output_features


class HipnnQuad(HipnnVec):
    """
    HIP-NN-TS with l=2
    """

    _interaction_class = InteractLayerQuad



"""
    Unit tests for ASE Graph construction. 
"""

def ASE_FilterPair_Coulomb_Construct():    
    """ Construct ASE Calculator from a  HIPNN model with multiple PairIndexers and Coulomb Energy Node. 
    Ensures that graph construction using PairFilter does not break ASE graph. 

    Returns:
        status : True if ASE calculator can be built. 
        exception : Caught Exception if ASE calculator build fails.
    """

    import numpy as np

    import torch
    torch.set_default_dtype(torch.float32)
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    torch.cuda.set_device(device)

    # Hippynn imports. 
    import hippynn
    from hippynn.graphs import inputs, networks, targets, physics
    from hippynn.graphs.nodes import indexers, pairs 

    nacl_species_set = [0, 11, 17] 
    network_params = {
        "possible_species": nacl_species_set,   # Z values of the elements
        'n_features': 10,                     # Number of neurons at each layer
        "n_sensitivities": 8,                # Number of sensitivity functions in an interaction layer
        "dist_soft_min": 0.90,  # qm7 1.7  qm9 .85  AL100 .85
        "dist_soft_max": 6.5,  # qm7 10.  qm9 5.   AL100 5.
        "dist_hard_max": 7.5,  # qm7 15.  qm9 7.5  AL100 7.5
        "n_interaction_layers": 1,            # Number of interaction blocks
        "n_atom_layers": 1,                   # Number of atom layers in an interaction bloc
    }

    ### Define Models 
    from hippynn.graphs import inputs, networks, targets, physics

    # Input node (should be shared across both networks)
    species = inputs.SpeciesNode(db_name="Z")
    positions = inputs.PositionsNode(db_name="R")

    ### ACA network 
    network_aca = networks.Hipnn("aca", (species, positions), module_kwargs=network_params)

    # Targets
    hcharge = targets.HChargeNode("HCharge", network_aca)
    atom_charges = hcharge.atom_charges 
    molecule_charges = physics.AtomToMolSummer("molCharge", atom_charges)
    dipole = physics.DipoleNode("dipole", (hcharge, positions), db_name="dipole")

    # Manually define index nodes for coulomb energy only!
    # coulomb_r_max should be large enough to cover all atoms. 
    # Large coulomb_r_max can lead to very slow training. 
    coulomb_r_max = 100 # np.inf -> errors for periodic systems. 

    enc, padidxer = indexers.acquire_encoding_padding(
        species, 
        species_set = nacl_species_set
    )
    pairfinder = pairs.OpenPairIndexer(
        'OpenPairFinder', 
        (positions, enc, padidxer), 
        dist_hard_max=coulomb_r_max
    )

    # Coulomb Energy
    energy_conv = 14.397 # Coulomb-konst
    coulomb_energy = physics.CoulombEnergyNode(
        "cEnergy",
        # _input_names = "charges", "pair_dist", "pair_first", "pair_second", "mol_index", "n_molecules"
        (atom_charges,
            pairfinder.pair_dist, pairfinder.pair_first, pairfinder.pair_second,
            padidxer.mol_index, padidxer.n_molecules),
        energy_conversion=energy_conv, 
    )
    ### 

    # Energy Network. 
    network_energy = networks.Hipnn("hipnn", network_aca.parents, module_kwargs=network_params)    
    Henergy = targets.HEnergyNode("HEnergy", network_energy) 
    henergy = Henergy.mol_energy + coulomb_energy 
    force = physics.GradientNode("Grad", (henergy, positions), sign=+1)
    henergy.db_name = "energy" 
    force.db_name = "gradient"


    ### Define Loss Graph
    from hippynn.graphs import loss 

    # All the ACA losses 
    rmse_dipole =  loss.MSELoss.of_node(dipole) ** (1/2)
    mae_dipole = loss.MAELoss.of_node(dipole)
    rsq_dipole = loss.Rsq.of_node(dipole)
    loss_charge = loss.MeanSq.of_node(molecule_charges)**(1/2) # Regularizer for net-charge = 0 
    l2_reg_aca = loss.l2reg(network_aca)
    loss_reg_aca = 1e-4 * (l2_reg_aca) 
    loss_dipole = rmse_dipole + mae_dipole
    loss_aca = loss_charge + loss_dipole + loss_reg_aca

    # All Energy Loss
    mse_force = loss.MSELoss.of_node(force)
    rmse_force = mse_force ** (1 / 2)
    mae_force = loss.MAELoss.of_node(force)
    rsq_force =  loss.Rsq.of_node(force)
    rmse_energy = loss.MSELoss.of_node(henergy) ** (1 / 2)
    mae_energy = loss.MAELoss.of_node(henergy)
    rsq_energy =  loss.Rsq.of_node(henergy)
    loss_energy = (rmse_energy + mae_energy)
    loss_force = (rmse_force + mae_force)
    loss_hipnn = loss_energy + loss_force
    loss_train = loss_aca + loss_hipnn

    # Validation losses
    validation_losses = {
        "T-RMSE"      : rmse_energy,
        "T-MAE"       : mae_energy,
        "T-RSQ"       : rsq_energy,
        "F-RMSE"      : rmse_force,
        "F-MAE"       : mae_force,
        "F-RSQ"       : rsq_force,
        "D-RMSE"      : rmse_dipole,
        "D-MAE"       : mae_dipole,
        "D-RSQ"       : rsq_dipole,
        "C-Loss"      : loss_charge,
        "L2-reg"      : l2_reg_aca,
        "Loss_aca"    : loss_aca,
        "Loss_hipnn"  : loss_hipnn,
        "Loss"        : loss_train,
    }
    ###

    #Assemble Pytorch Model that can be trained. 
    training_modules, db_info = hippynn.experiment.assemble_for_training(
        loss_train,
        validation_losses,
    )

    # Extract the un-trained model    
    model = training_modules.model

    # Construct a Calculator from model. 
    from hippynn.interfaces.ase_interface import HippynnCalculator
    energy_node = model.node_from_name('add(HEnergy.mol_energy, cEnergy.mol_energies)')
    try:
        calc = HippynnCalculator(energy=energy_node)
        return True, None 
    except Exception as ex:
        return False, ex
        

if __name__ == "__main__":

    import unittest
    
    class ASE_FilterPICoulombTest(unittest.TestCase):
        def runTest(self):
            status, ex = ASE_FilterPair_Coulomb_Construct()
            self.assertEqual(status, True, "ASE Calculator Build Fails with exception : {}".format(ex))

    unittest.main()


import os
import warnings
from collections import defaultdict, OrderedDict
import matplotlib.pyplot as plt

from .. import settings

DEFAULT_TIMEPLOT_KWARGS = {
    "train": {
        "c": "r",
        "ms": 3,
        "lw": 1,
    },
    "valid": {
        "c": "b",
        "lw": 1,
        "ms": 3,
    },
}
DEFAULT_TIMEPLOT_KWARGS = defaultdict(dict, DEFAULT_TIMEPLOT_KWARGS)


def plot_all_over_time(metric_list, best_metric_list, save_dir="plots/over_time"):
    transposed_metrics = defaultdict(lambda: defaultdict(list))
    transposed_best = defaultdict(lambda: defaultdict(list))
    for emet, ebest in zip(metric_list, best_metric_list):
        for eval_type, this_dict in emet.items():
            for metric_name, mval in this_dict.items():
                bmval = ebest[eval_type][metric_name]
                transposed_metrics[metric_name][eval_type].append(mval)
                transposed_best[metric_name][eval_type].append(bmval)

    if os.path.exists(save_dir):
        warnings.warn("May override plots in {} !".format(save_dir))
    else:
        os.makedirs(save_dir)

    for metric_name in transposed_metrics:
        this_met_data = OrderedDict()
        pltkw_info = dict()
        for type_name in transposed_metrics[metric_name]:
            this_met_data[type_name] = transposed_metrics[metric_name][type_name]
            pltkw_info[type_name] = {"ls": "-", "marker": "o"}
            pltkw_info[type_name].update(DEFAULT_TIMEPLOT_KWARGS[type_name])

            best_name = type_name + "-best"
            this_met_data[best_name] = transposed_best[metric_name][type_name]
            pltkw_info[best_name] = {"ls": ":"}
            pltkw_info[best_name].update(DEFAULT_TIMEPLOT_KWARGS[type_name])

        plot_over_time(metric_name, this_met_data, pltkw_info, save_dir=save_dir)


def plot_over_time(metric_name, metric_data, pltkwd_info, save_dir):
    fig, ax = plt.subplots()

    plt.title(metric_name)
    plt.xlabel("Epoch")
    plt.ylabel(metric_name)
    for key, data in metric_data.items():
        plt.plot(data, **pltkwd_info[key], label=key)
    plt.legend()
    fname = os.path.join(save_dir, metric_name + settings.DEFAULT_PLOT_FILETYPE)
    # TODO: Check fname against characters meaningful to the filesystem, e.g. `/`.
    plt.savefig(fname)

    datamin = min(min(arr) for arr in metric_data.values())
    datamax = max(max(arr) for arr in metric_data.values())
    if datamin > 0 and datamax / datamin > 2:
        plt.yscale("log")
        fname = os.path.join(save_dir, metric_name + "_logplot" + settings.DEFAULT_PLOT_FILETYPE)
        plt.savefig(fname)

        if max(len(arr) for arr in metric_data.values()) > 10:
            plt.xscale("log")
            fname = os.path.join(save_dir, metric_name + "_loglogplot" + settings.DEFAULT_PLOT_FILETYPE)
            plt.savefig(fname)

    plt.close(fig)



"""
This file is used to host the function `set_devices`, as it is needed by both
serialization and routines. We may find a better place for this function in the
future.
"""
from typing import Tuple, Union

import torch

from ..graphs import GraphModule
from .evaluator import Evaluator


def set_devices(
    model: GraphModule,
    loss: GraphModule,
    evaluator: Evaluator,
    optimizer: torch.optim.Optimizer,
    device: Union[torch.device, str, Tuple, list],
) -> Tuple:
    """Sets the model, loss, and optimizer to the specified device if necessary.
    Evaluation loss is performed on CPU.

    :param model: current model on CPU
    :type model: GraphModule
    :param loss: current loss module on CPU
    :type loss: GraphModule
    :param evaluator: evaluator
    :type evaluator: Evaluator
    :param optimizer: optimizer with state dictionary on CPU
    :type optimizer: torch.optim.Optimizer
    :param device: target device for training. If device is a tuple or list, wrap
        the model using torch.nn.DataParallel and use the first specified device
        as the "primary: one.
    :type device: Union[torch.devices, string, tuple, list]
    :return: model, evaluator, optimizer
    :rtype: Tuple
    """
    print("Using device: ", device)
    if isinstance(device, (tuple, list)):
        model = torch.nn.DataParallel(model, device_ids=device)
        print("Using multi GPU compute on devices:", device)
        device = torch.device(device[0])
    device = torch.device(device)  # Will throw a more explicit error if the device specification is invalid

    model.to(device)
    loss.to(device)
    evaluator.loss.cpu()
    evaluator.model_device = device
    evaluator.model = model
    # reload the state dict after model transfer
    optimizer.load_state_dict(optimizer.state_dict())

    return model, evaluator, optimizer



import itertools

import matplotlib
from matplotlib import pyplot as plt

from ..graphs.indextypes.reduce_funcs import elementwise_compare_reduce
from .. import settings

import torch


class Plotter:
    """
    Base class for plotters.
    Can be used with external callbacks by specifying plt_fn.
    """

    def __init__(self, parents, plt_fn=None, saved=False, shown=False):
        self.parents = parents
        self.shown = shown
        self.saved = saved
        if callable(plt_fn):
            self.plt_fn = plt_fn

    def make_plot(self, data_args):
        fig, ax = plt.subplots()
        self.plt_fn(*data_args)
        return fig

    def plt_fn(self):
        return NotImplemented


def as_numpy(torch_tensor):
    return torch_tensor.data.cpu().numpy()


class ComposedPlotter(Plotter):
    def __init__(self, subplotters, **kwargs):
        self.subplotters = subplotters
        parents = list(itertools.chain.from_iterable((p.parents for p in subplotters)))
        super().__init__(parents, **kwargs)

    def plt_fn(self, *data_args):
        remaining_args = data_args
        for plotter in self.subplotters:
            n_args = len(plotter.parents)
            this_args = remaining_args[:n_args]
            remaining_args = remaining_args[n_args:]
            fig = plotter.plt_fn(*this_args)
        return fig


class _CompareableTruePred:
    @classmethod
    def compare(cls, node, **kwargs):
        node = node.main_output
        xlabel = "true " + node.db_name
        ylabel = "predicted " + node.db_name
        if kwargs.get("saved", False) is True:
            kwargs["saved"] = cls.__name__ + "-" + node.db_name + settings.DEFAULT_PLOT_FILETYPE
        reduced_true, reduced_pred = elementwise_compare_reduce(node.true, node.pred)
        return cls(reduced_true, reduced_pred, xlabel=xlabel, ylabel=ylabel, **kwargs)


class Hist2D(Plotter, _CompareableTruePred):
    def __init__(self, x_var, y_var, xlabel, ylabel, bins=200, norm=None, add_identity_line=True, **kwargs):
        super().__init__((x_var, y_var), **kwargs)
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.bins = bins
        self._norm = norm
        self.add_identity_line = add_identity_line

    def plt_fn(self, x_val, y_val):
        x_val = as_numpy(x_val).flatten()
        y_val = as_numpy(y_val).flatten()
        plt.hist2d(x_val, y_val, bins=self.bins, norm=self.norm)
        if self.add_identity_line:
            min_val = min(x_val.min(), y_val.min())
            max_val = max(x_val.max(), y_val.max())
            plt.plot((min_val, max_val), (min_val, max_val), c="r", lw=0.5)
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)
        plt.colorbar()

    @property
    def norm(self):
        return self._norm or matplotlib.colors.LogNorm()


class HierarchicalityPlot(Hist2D):
    def __init__(self, x_var, y_var, **kwargs):
        super().__init__(x_var, y_var, xlabel="log_10 R", ylabel="|ΔE|", add_identity_line=False, **kwargs)

    def plt_fn(self, x_val, y_val):
        with torch.autograd.no_grad():
            x_val = torch.log10(x_val + 1e-15)
            y_val = torch.abs(y_val)
        return super().plt_fn(x_val, y_val)


class Hist1DComp(Plotter, _CompareableTruePred):
    def __init__(self, x_var, y_var, xlabel, ylabel, bins=200, norm=None, **kwargs):
        super().__init__((x_var, y_var), **kwargs)
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.bins = bins
        self._norm = norm

    def plt_fn(self, x_val, y_val):
        x_val = as_numpy(x_val).flatten()
        y_val = as_numpy(y_val).flatten()
        plt.hist(x_val, bins=self.bins, label=self.xlabel, color=[1.0, 0, 0, 0.5])
        plt.hist(y_val, bins=self.bins, label=self.ylabel, color=[0, 0, 1.0, 0.5])
        plt.xlabel(self.xlabel + " and " + self.ylabel)
        plt.ylabel("Count")
        plt.legend()


class Hist1D(Plotter):
    def __init__(self, x_var, xlabel, bins=200, **kwargs):
        super().__init__((x_var,), **kwargs)
        self.xlabel = xlabel
        self.bins = bins

    def plt_fn(self, x_val):
        plt.hist(x_val, bins=self.bins)
        plt.xlabel(self.xlabel)
        plt.ylabel("Count")


class SensitivityPlot(Plotter):
    def __init__(self, sensitivity_module, r_min=0.1, r_max=None, n_r=500, **kwargs):
        self.sensitivity = sensitivity_module
        if r_max is None:
            r_max = self.sensitivity.hard_max_dist + 1
        self.r_params = r_min, r_max, n_r
        super().__init__((), **kwargs)

    def plt_fn(self):
        with torch.autograd.no_grad():
            mu = list(self.sensitivity.parameters())[0]
            r_range = torch.linspace(*self.r_params, dtype=mu.dtype, device=mu.device)
            # allow_warning=False to disable the false low distance warning
            sense_out = self.sensitivity(r_range, warn_low_distances=False).cpu().data.numpy()
            r_range = r_range.cpu().data.numpy()
        for sense_func in sense_out.transpose():
            plt.plot(r_range, sense_func, c="r")
        plt.plot(r_range, sense_out.sum(axis=1), c="b")
        plt.xlabel("Distance")
        plt.ylabel("Sensitivity")


class InteractionPlot(Plotter):
    def __init__(self, int_layer, r_min=0.1, r_max=None, n_r=500, **kwargs):
        raise NotImplementedError()  #  TODO Finish this.

        self.int_layer = int_layer
        if r_max is None:
            r_max = self.int_layer.sensitivity.hard_max_dist + 0.5
        self.r_params = r_min, r_max, n_r
        super().__init__((), **kwargs)

    def plt_fn(self):
        dtype, device = self.int_layer.int_weights.dtype, self.int_layer.int_weights.device



"""
Nodes for constructing loss functions.
"""
import torch
import torch.nn.functional

from ... import settings
from ..indextypes import IdxType, elementwise_compare_reduce
from .base import SingleNode
from ...layers.algebra import LambdaModule

from ...layers import regularization as reg_modules


class _DebugBroadCast(torch.nn.Module):
    def __init__(self, inner):
        super().__init__()
        self.inner = inner

    def extra_repr(self):
        return repr(self.inner)

    def forward(self, *inputs):
        shapes = [x.shape for x in inputs]
        print("DEBUG BROADCASTING shapes:", shapes)
        for s in shapes:
            if not (s == shapes[0]):
                raise ValueError("Broadcasting in loss shapes:", shapes)
        return self.inner(*inputs)


class ReduceSingleNode(SingleNode):
    _index_state = IdxType.Scalar

    def __init__(self, parent):
        name = self._classname + "({})".format(parent.name)
        parent = elementwise_compare_reduce(parent.main_output)
        super().__init__(name, (parent,))

    @classmethod
    def of_node(cls, node):
        return cls(node.pred)

    def __init_subclass__(cls, op=None, **kwargs):
        if op is not None:
            cls._classname = op.__name__
            cls.torch_module = LambdaModule(op)


class Mean(ReduceSingleNode, op=torch.mean):
    pass


def mean_sq(input):
    return torch.pow(input, 2).mean()


class MeanSq(ReduceSingleNode, op=mean_sq):
    pass


class Std(ReduceSingleNode, op=torch.std):
    pass


class Var(ReduceSingleNode, op=torch.var):
    pass


class _BaseCompareLoss(SingleNode):
    _index_state = IdxType.Scalar

    def __init__(self, predicted, true):
        name = "{}({},{})".format(self._classname, predicted.name, true.name)
        predicted, true = elementwise_compare_reduce(predicted, true)
        super().__init__(name, (predicted, true), module=None)

    @classmethod
    def of_node(cls, node):
        node = node.main_output
        return cls(node.pred, node.true)

    def __init_subclass__(cls, op=None, **kwargs):
        if op is not None:
            cls._classname = op.__name__
            if settings.DEBUG_LOSS_BROADCAST:
                cls.torch_module = _DebugBroadCast(op)
            else:
                # Note: as of now, we need to wrap raw operations as a loss module because
                # the graph module's ModuleList can't take non-torch operations.
                # Add the lambda to the graph module instead to put the problem nearer to the solution?
                cls.torch_module = LambdaModule(op)


class RsqMod(torch.nn.Module):
    def forward(self, predicted, true):
        return 1 - (torch.mean(torch.pow(predicted - true, 2)) / true.var())


class Rsq(_BaseCompareLoss):
    torch_module = RsqMod()
    _classname = "Rsq"


class MSELoss(_BaseCompareLoss, op=torch.nn.functional.mse_loss):
    pass


class MAELoss(_BaseCompareLoss, op=torch.nn.functional.l1_loss):
    pass


class _LPReg(SingleNode):
    _index_state = IdxType.Scalar
    _auto_module_class = reg_modules.LPReg

    def __init__(self, network, p=2, module="auto"):
        name = "L^P_Reg({},p={})".format(network.name, p)
        parents = (network,)
        self.p = p
        super().__init__(name, parents, module=module)

    def auto_module(self):
        return self._auto_module_class(self.parents[0].torch_module, p=self.p)


def lpreg(network, p):
    return _LPReg(network, p=p).pred


def l2reg(network):
    return lpreg(network, p=2)


def l1reg(network):
    return lpreg(network, p=1)

# For loss functions with phases
def absolute_errors(predict: torch.Tensor, true: torch.Tensor):
    """Compute the absolute errors with phases between predicted and true values. In
    other words, prediction should be close to the absolute value of true, and the sign
    does not matter.

    :param predict: predicted values
    :type predict: torch.Tensor
    :param true: true values
    :type true: torch.Tensor
    :return: errors
    :rtype: torch.Tensor
    """

    return torch.minimum(torch.abs(true - predict), torch.abs(true + predict))





"""
Nodes for prediction of variables from network features.
"""

from .base import MultiNode, AutoKw, ExpandParents, find_unique_relative, _BaseNode
from .tags import AtomIndexer, Network, PairIndexer, HAtomRegressor, Charges, Energies
from .indexers import PaddingIndexer
from ..indextypes import IdxType, index_type_coercion
from ...layers import targets as target_modules


class HEnergyNode(Energies, HAtomRegressor, AutoKw, ExpandParents, MultiNode):
    """
    Predict a system-level scalar such as energy from a sum over local components.
    """

    _input_names = "hier_features", "mol_index", "n_molecules"
    _output_names = "mol_energy", "atom_energies", "energy_terms", "hierarchicality", "atom_hier", "mol_hier", "batch_hier"
    _main_output = "mol_energy"
    _output_index_states = IdxType.Molecules, IdxType.Atoms, None, IdxType.Molecules, IdxType.Atoms, IdxType.Molecules, IdxType.Scalar
    _auto_module_class = target_modules.HEnergy

    @_parent_expander.match(Network)
    def expansion0(self, net, **kwargs):
        if "feature_sizes" not in self.module_kwargs:
            self.module_kwargs["feature_sizes"] = net.torch_module.feature_sizes
        pdindexer = find_unique_relative(net, AtomIndexer)
        return net, pdindexer.mol_index, pdindexer.n_molecules

    def __init__(self, name, parents, first_is_interacting=False, module="auto", module_kwargs=None, **kwargs):
        self.module_kwargs = {"first_is_interacting": first_is_interacting}
        if module_kwargs is not None:
            self.module_kwargs = {**self.module_kwargs, **module_kwargs}
        parents = self.expand_parents(parents, **kwargs)
        super().__init__(name, parents, module=module, **kwargs)


class HChargeNode(Charges, HAtomRegressor, AutoKw, ExpandParents, MultiNode):
    """
    Predict an atom-level scalar such as charge from local features.
    """

    _input_names = ("hier_features",)
    _output_names = "atom_charges", "partial_sums", "charge_hierarchality"
    _main_output = "atom_charges"
    _output_index_states = IdxType.Atoms, None, IdxType.Atoms
    _auto_module_class = target_modules.HCharge

    @_parent_expander.match(Network)
    def expansion0(self, net, **kwargs):
        if "feature_sizes" not in self.module_kwargs:
            self.module_kwargs["feature_sizes"] = net.torch_module.feature_sizes
        return (net.main_output,)

    def __init__(self, name, parents, module="auto", module_kwargs=None, **kwargs):
        self.module_kwargs = {} if module_kwargs is None else module_kwargs
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


class LocalChargeEnergy(Energies, ExpandParents, HAtomRegressor, MultiNode):
    _input_names = "charges", "features", "mol_index", "n_molecules"
    _output_names = "molenergies", "atomenergies"
    _main_output = "molenergies"
    _output_index_states = IdxType.Molecules, IdxType.Atoms
    _auto_module_class = target_modules.LocalChargeEnergy

    @_parent_expander.match(_BaseNode, Network)
    def expansion0(self, charge, network, **kwargs):
        charge = index_type_coercion(charge.main_output, IdxType.Atoms)
        pdxer = find_unique_relative(network, PaddingIndexer)
        return charge, network.main_output, pdxer.mol_index, pdxer.n_molecules

    def __init__(self, name, parents, **kwargs):
        parents = self.parents_expand(parents)
        super().__init__(name, parents, **kwargs)


class HBondNode(ExpandParents, AutoKw, MultiNode):
    """
    Predict an pair-level scalar such as bond order from local features on both atoms
    """

    _auto_module_class = target_modules.HBondSymmetric
    _output_names = "bonds", "bond_hierarchality"
    _output_index_states = IdxType.Pair, IdxType.Pair
    _input_names = "features", "pair_first", "pair_second", "pair_dist"
    _main_output = "bonds"
 
    @_parent_expander.matchlen(1)
    def expand0(self, features, *, purpose, **kwargs):
        pairfinder = find_unique_relative(features, PairIndexer, why_desc=purpose)
        return features, pairfinder

    @_parent_expander.matchlen(2)
    def expand1(self, features, pairfinder, **kwargs):
        return features.main_output, pairfinder.pair_first, pairfinder.pair_second, pairfinder.pair_dist

    def __init__(self, name, parents, module="auto", module_kwargs=None, **kwargs):
        self.module_kwargs = {} if module_kwargs is None else module_kwargs
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)





"""
Nodes for indexing information.
"""
from .tags import Encoder, AtomIndexer
from .base import SingleNode, AutoNoKw, AutoKw, find_unique_relative, MultiNode, ExpandParents, _BaseNode
from .base.node_functions import NodeNotFound
from .inputs import SpeciesNode

# Index generating functions need access to appropriately raise this
from ..indextypes import IdxType
from ...layers import indexers as index_modules


class OneHotEncoder(AutoKw, Encoder, MultiNode):
    """
    Node for encoding species as one-hot vectors
    """

    _output_names = "encoding", "nonblank"
    _output_index_states = IdxType.MolAtom, IdxType.MolAtom
    _auto_module_class = index_modules.OneHotSpecies

    def __init__(self, name, parents, species_set, module="auto", **kwargs):
        try:
            species_set = species_set.clone()
        except AttributeError:
            pass  # If was not passed a tensor.
        self.species_set = species_set

        # Can be passed just a single species node, not a tuple
        if isinstance(parents, _BaseNode):
            parents = (parents,)
        super().__init__(name, parents, module=module, **kwargs)

    def auto_module(self):
        return self._auto_module_class(species_set=self.species_set)


class PaddingIndexer(AtomIndexer, AutoNoKw, ExpandParents, MultiNode):
    """
    Node for building information to convert from
    MolAtom to Atom index state.
    """

    _output_names = (
        "indexed_features",
        "real_atoms",
        "inv_real_atoms",
        "mol_index",
        "atom_index",
        "n_molecules",
        "n_atoms_max",
    )
    _output_index_states = IdxType.Atoms, None, None, None, None, None, None  # optional?
    _input_names = "encoding", "nonblank"
    _auto_module_class = index_modules.PaddingIndexer

    @_parent_expander.match(Encoder)
    def expand0(self, encoder, **kwargs):
        return encoder.encoding, encoder.nonblank

    _parent_expander.assertlen(2)

    def __init__(self, name, parents, *args, **kwargs):
        parents = self.expand_parents(parents)
        super().__init__(name, parents, *args, **kwargs)


class AtomReIndexer(ExpandParents, AutoNoKw, SingleNode):
    """
    Node for re-using index information to convert MolAtom->Atom.
    """

    _auto_module_class = index_modules.AtomReIndexer
    _index_state = IdxType.Atoms

    @_parent_expander.match(SingleNode)
    def expand0(self, features, *, purpose, **kwargs):
        pad_idx = find_unique_relative(features, PaddingIndexer, why_desc=purpose)
        return features, pad_idx

    @_parent_expander.match(SingleNode, PaddingIndexer)
    def expand1(self, features, pad_idx, **kwargs):
        return features, pad_idx.real_atoms

    _parent_expander.assertlen(2)

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


class AtomDeIndexer(ExpandParents, AutoNoKw, SingleNode):
    """
    Node for converting Atom->MolAtom
    """

    _auto_module_class = index_modules.AtomDeIndexer
    _index_state = IdxType.MolAtom

    @_parent_expander.matchlen(1)
    def expand0(self, features, *, purpose, **kwargs):
        pad_idx = find_unique_relative(features, PaddingIndexer, why_desc=purpose)
        return features, pad_idx.mol_index, pad_idx.atom_index, pad_idx.n_molecules, pad_idx.n_atoms_max

    @_parent_expander.matchlen(2)
    def expand0(self, features, mol_index, atom_index, n_mol, n_atom, **kwargs):
        return features.main_output, mol_index, atom_index, n_mol, n_atom

    _parent_expander.assertlen(5)

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


class QuadUnpackNode(AutoNoKw, SingleNode):
    _auto_module_class = index_modules.QuadUnpack
    _index_state = IdxType.Molecules

    def __init__(self, name, parents, module="auto", **kwargs):
        super().__init__(name, parents, module=module, **kwargs)


class FilterBondsOneway(AutoNoKw, SingleNode):
    """
    Node which filters the set of pairs to a one-way list.
    """

    _input_names = "input_bonds", "pair_first", "pair_second"
    _index_state = IdxType.NotFound
    _auto_module_class = index_modules.FilterBondsOneway

    def __init__(self, name, parents, module="auto", **kwargs):
        super().__init__(name, parents, module=module, **kwargs)


class SysMaxOfAtomsNode(ExpandParents, AutoNoKw, SingleNode):
    _input_names = "var", "mol_index", "n_molecules"
    _index_state = IdxType.Molecules
    _auto_module_class = index_modules.SysMaxOfAtoms

    @_parent_expander.match(_BaseNode)
    def expansion0(self, node, *, purpose, **kwargs):
        pidxer = find_unique_relative(node, AtomIndexer, why_desc=purpose)
        return node, pidxer

    @_parent_expander.match(_BaseNode, AtomIndexer)
    def expansion1(self, node, pidxer, *, purpose, **kwargs):
        return node, pidxer.mol_index, pidxer.n_molecules

    _parent_expander.assertlen(3)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.Atoms, None, None)

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


def acquire_encoding_padding(search_nodes, species_set, purpose=None):
    """
    Automatically finds/builds a one-hot encoder and padding indexer starting from ``search_nodes``.

    If the encoder and padder exist as relatives of the search nodes.

    :param search_nodes: Node or nodes to start from.
    :param species_set: Species set to use if an encoder needs to be created.
    :param purpose: String for error information if the process fails. (optional)
    :return: encoder, padding indexer

    """
    try:
        encoder = find_unique_relative(search_nodes, Encoder, why_desc=purpose)
    except NodeNotFound:
        if species_set is None:
            raise ValueError(
                "Building encode and padder requires a species_set, but a species set is not specified,"
                " Make an encoder for the needed species node e.g. using the command: \n"
                "`encoder = OneHotEncoder('OneHot', species_node, species_set=species_set)`.\n"
            )
        species_node = find_unique_relative(search_nodes, SpeciesNode, why_desc=purpose)
        encoder = OneHotEncoder("OneHot", (species_node,), species_set=species_set)

    try:
        pidxer = find_unique_relative(search_nodes, PaddingIndexer, why_desc=purpose)
    except NodeNotFound:
        pidxer = PaddingIndexer("PaddingIndexer", (encoder.encoding, encoder.nonblank))

    return encoder, pidxer

class FuzzyHistogrammer(AutoKw, SingleNode):
    """ 
    Node for transforming a scalar feature into a vectorized feature via 
    the fuzzy/soft histogram method.

    :param length: length of vectorized feature
    """

    _input_names = "values"
    _auto_module_class = index_modules.FuzzyHistogram

    def __init__(self, name, parents, length, vmin, vmax, module="auto", **kwargs):
        
        if isinstance(parents, _BaseNode):
            parents = (parents,)

        self._output_index_state = parents[0]._index_state
        self.module_kwargs = {"length": length, "vmin": vmin, "vmax": vmax}

        super().__init__(name, parents, module=module, **kwargs)


"""
This module implements a version of converting

Note: This is not entirely API safe as numba has
not exposed all of these functions directly.
"""
import numpy as np

from numba.cuda.cudadrv import devices, devicearray
try:
    from numba.cuda.api_util import prepare_shape_strides_dtype
except ImportError:
    # older versions of numba
    from numba.cuda.api import _prepare_shape_strides_dtype as prepare_shape_strides_dtype

from numba.cuda.cudadrv import driver
from numba.core import config

require_context = devices.require_context
current_context = devices.get_context
gpus = devices.gpus

import torch

typedict = {
    torch.complex64: "<c8",
    torch.complex128: "<c16",
    torch.float16: "<f2",
    torch.float32: "<f4",
    torch.float64: "<f8",
    torch.uint8: "|u1",
    torch.int8: "|i1",
    torch.int16: "<i2",
    torch.int32: "<i4",
    torch.int64: "<i8",
}
typedict = {k:np.dtype(typestr) for k,typestr in typedict.items()}

@require_context
def batch_convert_torch_to_numba(*tensors,typedict=typedict):
    out = []

    for tensor in tensors:
        #v2 array interface only

        # CUDA devices are little-endian and tensors are stored in native byte
        # order. 1-byte entries are endian-agnostic.
        dtype = typedict[tensor.dtype]

        shape = tuple(tensor.shape)
        if tensor.is_contiguous():
            # __cuda_array_interface__ v2 requires the strides to be omitted
            # (either not set or set to None) for C-contiguous arrays.
            strides = None
        else:
            itemsize = tensor.storage().element_size()
            strides = tuple(s * itemsize for s in tensor.stride())
        data_ptr = tensor.data_ptr() if tensor.numel() > 0 else 0
        data = (data_ptr, False)  # read-only is false


        shape = shape
        strides = strides

        shape, strides, dtype = prepare_shape_strides_dtype(
            shape, strides, dtype, order='C')
        size = driver.memory_size_from_info(shape, strides, dtype.itemsize)
        devptr = driver.get_devptr_for_active_ctx(data_ptr)
        data = driver.MemoryPointer(
            current_context(), devptr, size=size, owner=tensor)
        stream = 0  # No "Numba default stream", not the CUDA default stream
        da = devicearray.DeviceNDArray(shape=shape, strides=strides,
                                       dtype=dtype, gpu_data=data,
                                       stream=stream)
        out.append(da)

    return out



import torch
from hippynn.graphs import loss
from hippynn.experiment.serialization import load_checkpoint
from hippynn.experiment.assembly import assemble_for_training
from hippynn.graphs.nodes.inputs import PositionsNode, SpeciesNode
from .seqm_modules import pack_par

# hippynn.custom_kernels.set_custom_kernels (False)

if torch.cuda.is_available():
    device = torch.device("cuda")
else:
    device = torch.device("cpu")


class gen_par(torch.nn.Module):
    def __init__(
        self,
        model_file="experiment_structure.pt",
        state_file="best_checkpoint.pt",
        par_atom_node_name="SEQM_Atom_Params",
        seqm_node_name="SEQM_Energy",
        device=device,
    ):
        super().__init__()
        self.model_file = model_file
        self.state_file = state_file
        self.par_atom_node_name = par_atom_node_name
        self.seqm_node_name = seqm_node_name
        structure = torch.load(model_file, map_location=device)
        state = torch.load(state_file, map_location=device)
        structure["training_modules"][0].load_state_dict(state["model"])
        structure["controller"].load_state_dict(state["controller"])
        self.model = structure["training_modules"][0]
        if isinstance(self.model, torch.nn.DataParallel):
            self.model = self.model.module
        self.model.to(device)
        self.model.eval()
        par_atom = self.model.node_from_name(par_atom_node_name)
        seqm_par = par_atom.atom_charges
        train_loss = loss.MeanSq(seqm_par.pred)
        validation_losses = {"par_atom": train_loss}
        self.model_par = assemble_for_training(train_loss, validation_losses)[0][0]
        self.model_par.eval()
        if isinstance(self.model_par.input_nodes[0], SpeciesNode):
            self.input_order = True
        else:
            self.input_order = False
        self.seqm_module = self.model.node_from_name(seqm_node_name).torch_module

    def forward(self, species, positions):
        if self.input_order:
            input = [species, positions]
        else:
            input = [positions, species]
        par = self.model_par(*input)[0]
        learned_parameters = pack_par(self.seqm_module, species, par)
        return learned_parameters



"""
Machinery for nodes to support native python operators
such as addition, multiplication, subtraction.
"""
import functools
import operator

from ...indextypes import IdxType, get_reduced_index_state, elementwise_compare_reduce
from .node_functions import _BaseNode
from ....layers import algebra as algebra_mods


def wrap_as_node(obj):
    return obj.main_output if isinstance(obj, _BaseNode) else ValueNode(obj)


def coerces_values_to_nodes(func):
    """Wraps non-nodes as ValueNodes."""

    @functools.wraps(func)
    def newfunc(*args):
        return func(*(wrap_as_node(a) for a in args))

    return newfunc


class _CombNode(_BaseNode):
    @coerces_values_to_nodes
    def __add__(self, other):
        return AddNode(self, other)

    @coerces_values_to_nodes
    def __sub__(self, other):
        return SubNode(self, other)

    @coerces_values_to_nodes
    def __mul__(self, other):
        return MulNode(self, other)

    @coerces_values_to_nodes
    def __truediv__(self, other):
        return DivNode(self, other)

    @coerces_values_to_nodes
    def __pow__(self, other):
        return PowNode(self, other)

    @coerces_values_to_nodes
    def __radd__(self, other):
        return AddNode(other, self)

    @coerces_values_to_nodes
    def __rsub__(self, other):
        return SubNode(other, self)

    @coerces_values_to_nodes
    def __rmul__(self, other):
        return MulNode(other, self)

    @coerces_values_to_nodes
    def __rtruediv__(self, other):
        return DivNode(other, self)

    @coerces_values_to_nodes
    def __rpow__(self, other):
        return DivNode(other, self)

    def __invert__(self):
        return InvNode(self)

    def __pos__(self):
        return self

    def __neg__(self):
        return NegNode(self)


class ValueNode(_CombNode):
    _index_state = IdxType.Scalar

    def __init__(self, value):
        name = "Value({})".format(str(value))
        self.value = value
        super().__init__(name, parents=(), module="auto")

    def auto_module(self):
        return algebra_mods.ValueMod(self.value)


class _PredefinedOp:
    def __init_subclass__(cls, *, op=None, **kwargs):
        super().__init_subclass__(**kwargs)
        if op is not None:
            cls.torch_module = algebra_mods.LambdaModule(op)
            cls._classname = op.__name__


class UnaryNode(_PredefinedOp, _CombNode):
    def __init__(self, in_node):
        name = "{}({})".format(self._classname, in_node)
        super().__init__(name, (in_node,), module=None)
        self._index_state = in_node._index_state


class InvNode(UnaryNode, op=operator.invert):
    pass


class NegNode(UnaryNode, op=operator.neg):
    pass


class BinNode(_PredefinedOp, _CombNode):
    _classname = None

    def __init__(self, left, right):
        left, right = left.main_output, right.main_output
        idxstate = get_reduced_index_state(left, right)
        left, right = elementwise_compare_reduce(left, right)
        name = "{}({}, {})".format(self._classname, left.name, right.name)
        super().__init__(name, (left, right), module=None)
        self._index_state = idxstate


class AddNode(BinNode, op=operator.add):
    pass


class SubNode(BinNode, op=operator.sub):
    pass


class MulNode(BinNode, op=operator.mul):
    pass


class DivNode(BinNode, op=operator.truediv):
    pass


class PowNode(BinNode, op=operator.pow):
    pass


# This Node exists to prevent potential broadcasting problems, for example in the loss.
# Model-based quantities all use a feature index, even if the size is 1,
# e.g. energy is predicted with shape (n_molecules, 1)
# This AtLeast2D is then used to wrap things coming from the database so that they will
# have at least two dimensions.
# See nodes/loss.py and turn on `debug_loss_broadcast` if you have concerns about
# broadcasting behavior.
class AtLeast2D(_BaseNode):
    torch_module = algebra_mods.AtLeast2D()
    _index_state = IdxType.NotFound

    def __init__(self, parents, *args, **kwargs):
        if len(parents) != 1:
            raise ValueError("AtLeast2D can only have 1 parent, got {}".format(len(parents)))
        p = parents[0]
        self._index_state = p._index_state
        super().__init__("Atleast2D({})".format(p), parents, *args, module=None, **kwargs)
        self.origin_node = p.origin_node



"""
Nodes for physics transformations
"""
import warnings

from ...layers import indexers as index_layers
from ...layers import pairs as pair_layers
from ...layers import physics as physics_layers
from ..indextypes import IdxType, elementwise_compare_reduce, index_type_coercion
from .base import (
    AutoKw,
    AutoNoKw,
    ExpandParents,
    MultiNode,
    SingleNode,
    _BaseNode,
    find_unique_relative,
)
from .base.node_functions import NodeNotFound
from .indexers import AtomIndexer, PaddingIndexer, acquire_encoding_padding
from .inputs import PositionsNode, SpeciesNode
from .pairs import OpenPairIndexer
from .tags import Charges, Encoder, Energies, PairIndexer


class GradientNode(AutoKw, SingleNode):
    """
    Compute the gradient of a quantity.
    """

    _input_names = "energy", "coordinates"
    _auto_module_class = physics_layers.Gradient

    def __init__(self, name, parents, sign, module="auto", **kwargs):
        self.module_kwargs = {"sign": sign}
        energy, position = parents
        position.requires_grad = True
        parents = energy.main_output, position
        self.sign = sign
        self._index_state = position._index_state
        super().__init__(name, parents, module=module, **kwargs)
        
class MultiGradientNode(AutoKw, MultiNode):
    """
    Compute the gradient of a quantity.
    """

    _auto_module_class = physics_layers.MultiGradient

    def __init__(self, name: str, molecular_energies_parent: _BaseNode, generalized_coordinates_parents: tuple[_BaseNode], signs: tuple[int], module="auto", **kwargs):
        if isinstance(signs, int):
            signs = (signs,)

        self.signs = signs
        self.module_kwargs = {"signs": signs}

        parents = molecular_energies_parent, *generalized_coordinates_parents

        for parent in generalized_coordinates_parents:
            parent.requires_grad = True        

        self._input_names = tuple((parent.name for parent in parents))
        self._output_names = tuple((parent.name + "_grad" for parent in generalized_coordinates_parents))
        self._output_index_states = tuple(parent._index_state for parent in generalized_coordinates_parents)

        super().__init__(name, parents, module=module, **kwargs)


class StressForceNode(AutoNoKw, MultiNode):
    _input_names = "energy", "strain", "coordinates", "cell"
    _output_names = "forces", "stress"
    _auto_module_class = physics_layers.StressForce

    def __init__(self, name, parents, module="auto", **kwargs):
        energy, strain, coordinates, cell = parents
        coordinates.requires_grad = True
        parents = energy.main_output, strain, coordinates, cell
        self._output_index_states = coordinates._index_state, strain._index_state
        super().__init__(name, parents, module=module, **kwargs)


class ChargeMomentNode(ExpandParents, AutoNoKw, SingleNode):
    _input_names = "charges", "positions", "mol_index", "n_molecules"

    @_parent_expander.matchlen(1)
    def expansion0(self, charges, *, purpose, **kwargs):
        return charges, find_unique_relative(charges, PositionsNode, why_desc=purpose)

    @_parent_expander.match(Charges, PositionsNode)
    def expansion1(self, charges, positions, *, purpose, **kwargs):
        enc, pidxer = acquire_encoding_padding((charges, positions), species_set=None, purpose=purpose)
        return charges, positions, pidxer

    @_parent_expander.match(Charges, PositionsNode, AtomIndexer)
    def expansion2(self, charges, positions, pdxer, **kwargs):
        return charges, positions, pdxer.mol_index, pdxer.n_molecules

    _parent_expander.assertlen(4)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.Atoms, IdxType.Atoms, None, None)

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


class DipoleNode(ChargeMomentNode):
    """
    Compute the dipole of point charges.
    """

    _auto_module_class = physics_layers.Dipole
    _index_state = IdxType.Molecules


class QuadrupoleNode(ChargeMomentNode):
    """
    Compute the traceless quadrupole of point charges.
    """

    _auto_module_class = physics_layers.Quadrupole
    _index_state = IdxType.QuadMol


# Setup for Coulomb Energy and Screened Coulomb Energy is nearly the same, up to validating the pair finder.
class ChargePairSetup(ExpandParents):
    @staticmethod
    def _validate_pairfinder(pairfinder, cutoff_distance):
        # This method required by this ExpandParents setup.
        # Raises an error if the pairfinder is not satisfactory.
        return NotImplemented

    @_parent_expander.match(Charges)
    def expansion0(self, charges, *, purpose, **kwargs):
        try:
            pos_or_pair = find_unique_relative(charges, PairIndexer, why_desc=purpose)
        except NodeNotFound:
            pos_or_pair = find_unique_relative(charges, PositionsNode, why_desc=purpose)
        return charges, pos_or_pair

    @_parent_expander.match(Charges, PositionsNode)
    @_parent_expander.match(Charges, PairIndexer)
    def expansion1(self, charges, pos_or_pair, *, purpose, **kwargs):
        species = find_unique_relative((pos_or_pair, charges), SpeciesNode, why_desc=purpose)
        return charges, pos_or_pair, species

    @_parent_expander.match(Charges, SpeciesNode)
    def expansion1(self, charges, species, *, purpose, **kwargs):
        positions = find_unique_relative((charges, species), PositionsNode, why_desc=purpose)
        return charges, positions, species

    @_parent_expander.match(Charges, _BaseNode, SpeciesNode)
    def expansion2(self, charges, pos_or_pair, species, *, purpose, **kwargs):
        encoder, pidxer = acquire_encoding_padding(species, species_set=None, purpose=purpose)
        return charges, pos_or_pair, pidxer

    @_parent_expander.match(Charges, PositionsNode, PaddingIndexer)
    def expansion3(self, charges, positions, pidxer, *, cutoff_distance, **kwargs):
        try:
            pairfinder = find_unique_relative((charges, positions, pidxer), PairIndexer)
        except NodeNotFound:
            warnings.warn("Boundary conditions not specified, Building open boundary conditions.")
            encoder = find_unique_relative(pidxer, Encoder)
            pairfinder = OpenPairIndexer("PairIndexer", (positions, encoder, pidxer), dist_hard_max=cutoff_distance)
        return charges, pairfinder, pidxer

    @_parent_expander.match(Charges, PairIndexer, AtomIndexer)
    def expansion4(self, charges, pairfinder, pidxer, *, cutoff_distance, **kwargs):
        self._validate_pairfinder(pairfinder, cutoff_distance)
        pf = pairfinder
        return charges, pf.pair_dist, pf.pair_first, pf.pair_second, pidxer.mol_index, pidxer.n_molecules

    _parent_expander.assertlen(6)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.Atoms, *(None,) * 5)


class CoulombEnergyNode(ChargePairSetup, Energies, AutoKw, MultiNode):
    _input_names = "charges", "pair_dist", "pair_first", "pair_second", "mol_index", "n_molecules"
    _output_names = "mol_energies", "atom_energies", "atom_voltages"
    _output_index_states = IdxType.Molecules, IdxType.Atoms, IdxType.Atoms
    _main_output = "mol_energies"
    _auto_module_class = physics_layers.CoulombEnergy

    @staticmethod
    def _validate_pairfinder(pairfinder, cutoff_distance):
        if not isinstance(pairfinder, OpenPairIndexer):
            raise TypeError(
                "Closed boundary conditions detected.\n"
                "Coulomb energy module is not compatible with open boundary conditions."
            )

        if pairfinder.torch_module.hard_dist_cutoff is not None:
            raise ValueError(
                "hard_dist_cutoff is set to a finite value,\n"
                "coulomb energy requires summing over the entire set of pairs"
            )

    def __init__(self, name, parents, energy_conversion, module="auto"):
        """
        Besides the normal 'name' and 'parents' arguments, this node requires an `energy_conversion` parameter.
        This corresponds to coulomb's constant k in the equation E = kqq/r.
        """
        self.module_kwargs = {"energy_conversion_factor": energy_conversion}
        parents = self.expand_parents(parents, cutoff_distance=None)
        self.energy_conversion = energy_conversion
        super().__init__(name, parents, module=module)


class ScreenedCoulombEnergyNode(ChargePairSetup, Energies, AutoKw, MultiNode):
    _input_names = "charges", "pair_dist", "pair_first", "pair_second", "mol_index", "n_molecules"
    _output_names = "mol_energies", "atom_energies", "atom_voltages"
    _output_index_states = IdxType.Molecules, IdxType.Atoms, IdxType.Atoms
    _main_output = "mol_energies"
    _auto_module_class = physics_layers.ScreenedCoulombEnergy

    @staticmethod
    def _validate_pairfinder(pairfinder, cutoff_distance):
        existing_cutoff = pairfinder.torch_module.hard_dist_cutoff
        if existing_cutoff is not None and existing_cutoff < cutoff_distance:
            raise ValueError(
                f"Distance cutoff ({existing_cutoff}) is set to less than\n"
                f" pair finder distance ({cutoff_distance}). Increase the cutoff distance\n"
                f" for the pair_finder (named: {pairfinder.name})"
            )

    def __init__(self, name, parents, energy_conversion, cutoff_distance, screening=None, module="auto"):
        """
        Besides the normal 'name' and 'parents' arguments, this node requires an `energy_conversion` parameter.
        This corresponds to coulomb's constant k in the equation E = kqq/r.
        """

        if screening is None and module == "auto":
            raise ValueError(
                "To build this module automatically a screening module must\n"
                "be provided (e.g. layers.physiscs.QScreening(p_value=4))"
            )

        self.module_kwargs = {
            "energy_conversion_factor": energy_conversion,
            "radius": cutoff_distance,
            "screening": screening,
        }
        parents = self.expand_parents(
            parents,
            cutoff_distance=cutoff_distance,
        )
        self.energy_conversion = energy_conversion
        super().__init__(name, parents, module=module)


class VecMag(ExpandParents, AutoNoKw, SingleNode):
    _input_names = ("vector",)
    _auto_module_class = physics_layers.VecMag
    _index_state = IdxType.NotFound

    @_parent_expander.match(_BaseNode, _BaseNode)
    def expansion2(self, vector, helper, *, purpose, **kwargs):
        # This somewhat strange construction allows us to
        # find a padding indexer if the vector is detached from the padding indexer.
        vector, helper = elementwise_compare_reduce(vector, helper)
        return (vector,)

    _parent_expander.assertlen(1)
    _parent_expander.get_main_outputs()

    def __init__(self, name, parents, module="auto", _helper=None, **kwargs):
        parents = self.expand_parents(parents)
        self._index_state = parents[0]._index_state
        assert len(parents) == 1, "Improper number of parents for {}".format(self.__class__.__name__)
        super().__init__(name, parents, module=module, **kwargs)


class AtomToMolSummer(ExpandParents, AutoNoKw, SingleNode):
    _input_names = "features", "mol_index", "n_molecules"
    _auto_module_class = index_layers.MolSummer
    _index_state = IdxType.Molecules

    @_parent_expander.match(_BaseNode)
    def expansion0(self, features, **kwargs):
        pdxer = find_unique_relative(features, AtomIndexer, why_desc="Generating Molecular summer")
        return features, pdxer

    @_parent_expander.match(_BaseNode, AtomIndexer)
    def expansion1(self, features, pdxer, **kwargs):
        return features, pdxer.mol_index, pdxer.n_molecules

    _parent_expander.assertlen(3)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.Atoms, None, None)

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


# TODO: This seems broken for parent expanders, check the signature of the layer.
class BondToMolSummmer(ExpandParents, AutoNoKw, SingleNode):
    _input_names = "pairfeatures", "mol_index", "n_molecules", "pair_first"
    _auto_module_class = pair_layers.MolPairSummer
    _index_state = IdxType.Molecules

    @_parent_expander.match(_BaseNode)
    def expansion0(self, features, *, purpose, **kwargs):
        pdxer = find_unique_relative(features, AtomIndexer, why_desc=purpose)
        pair_idxer = find_unique_relative(features, PairIndexer, why_desc=purpose)
        return features, pdxer, pair_idxer

    @_parent_expander.match(_BaseNode, AtomIndexer, PairIndexer)
    def expansion1(self, features, pdxer, pair_idxer, **kwargs):
        return features, pdxer.mol_index, pdxer.n_molecules, pair_idxer.pair_first

    @_parent_expander.match(_BaseNode, _BaseNode, _BaseNode, _BaseNode, _BaseNode)
    def expansion2(self, features, mol_index, n_molecules, **kwargs):
        return index_type_coercion(features.main_output, IdxType.Pair), mol_index, n_molecules

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


class PerAtom(ExpandParents, AutoNoKw, SingleNode):
    _input_names = "features", "species"
    _index_state = IdxType.Molecules
    _auto_module_class = physics_layers.PerAtom

    @_parent_expander.match(_BaseNode)
    def expansion0(self, features, *, purpose, **kwargs):
        return features, find_unique_relative(features, SpeciesNode, purpose)

    @_parent_expander.match(_BaseNode, _BaseNode)
    def expansion1(self, features, species, **kwargs):
        features = features.main_output
        assert (
            features._index_state == IdxType.Molecules
        ), "Can only calculate Per Atom averages on Molecular quantities"
        return features, species

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


class CombineEnergyNode(Energies, AutoKw, ExpandParents, MultiNode):
    """
    Combines Local atom energies from different Energy Nodes.
    """

    _input_names = "input_atom_energy_1", "input_atom_energy_2", "mol_index", "n_molecules"
    _output_names = "mol_energy", "atom_energies"
    _main_output = "mol_energy"
    _output_index_states = (
        IdxType.Molecules,
        IdxType.Atoms,
    )
    _auto_module_class = physics_layers.CombineEnergy

    @_parent_expander.match(_BaseNode, Energies)
    def expansion0(self, energy_1, energy_2, **kwargs):
        return energy_1, energy_2.atom_energies

    @_parent_expander.match(Energies, _BaseNode)
    def expansion0(self, energy_1, energy_2, **kwargs):
        return energy_1.atom_energies, energy_2

    @_parent_expander.match(_BaseNode, _BaseNode)
    def expansion1(self, energy_1, energy_2, **kwargs):
        pdindexer = find_unique_relative([energy_1, energy_2], AtomIndexer, why_desc="Generating CombineEnergies")
        return energy_1, energy_2, pdindexer

    @_parent_expander.match(_BaseNode, _BaseNode, PaddingIndexer)
    def expansion2(self, energy_1, energy_2, pdindexer, **kwargs):
        return energy_1, energy_2, pdindexer.mol_index, pdindexer.n_molecules

    _parent_expander.assertlen(4)
    _parent_expander.require_idx_states(IdxType.Atoms, IdxType.Atoms, None, None)

    def __init__(self, name, parents, module="auto", module_kwargs=None, **kwargs):
        self.module_kwargs = {} if module_kwargs is None else module_kwargs
        parents = self.expand_parents(parents, **kwargs)
        super().__init__(name, parents=parents, module=module, **kwargs)



"""
Nodes for networks.
"""
from .tags import Encoder, PairIndexer, Network, AtomIndexer
from .base import _BaseNode, AutoKw, ExpandParents, SingleNode
from .indexers import OneHotEncoder, PaddingIndexer, acquire_encoding_padding
from .pairs import OpenPairIndexer, PeriodicPairIndexer
from .inputs import SpeciesNode, PositionsNode, CellNode
from ..indextypes import IdxType
from ... import networks as network_modules


class DefaultNetworkExpansion(ExpandParents):
    @_parent_expander.match(SpeciesNode, PositionsNode)
    @_parent_expander.match(SpeciesNode, PositionsNode, CellNode)
    def expansion0(self, species, *other_parents, species_set, purpose, **kwargs):
        """
        Finds or sets up a default one-hot encoder if species are passed as first argument.

        :return: (encoder,padding_indexer,*other_parents)

        """
        encoder, pidxer = acquire_encoding_padding(species, species_set, purpose=purpose)
        return (encoder, pidxer, *other_parents)

    @_parent_expander.match(Encoder, AtomIndexer, PositionsNode)
    @_parent_expander.match(Encoder, AtomIndexer, PositionsNode, CellNode)
    def expansion1(self, encoder, pidxer, positions, cell=None, *, dist_hard_max, periodic, **kwargs):
        """
        Setup pair finder if positions and cell are passed with encoding.

        :return: (padding_indexer, pairfinder)
        """
        if periodic:
            assert isinstance(cell, CellNode), "Periodic networks require a cell input"
            pair_parents = (positions, encoder, pidxer, cell)
            pair_cls = PeriodicPairIndexer
        else:
            assert cell is None, "When providing a cell node, periodic must be set to true"
            pair_parents = (positions, encoder, pidxer)
            pair_cls = OpenPairIndexer
        pairfinder = pair_cls("PairIndexer", pair_parents, dist_hard_max=dist_hard_max)
        return pidxer, pairfinder

    @_parent_expander.match(AtomIndexer, PairIndexer)
    def expansion1(self, pidxer, pairfinder, **kwargs):
        """
        Get indexed features from the atom indexer.

        :return: (indexed_features, pair_finder)
        """
        return pidxer.indexed_features, pairfinder


class Hipnn(DefaultNetworkExpansion, AutoKw, Network, SingleNode):
    """
    Node for HIP-NN neural networks
    """

    _input_names = "input_features", "pair_first", "pair_second", "pair_dist"
    _index_state = IdxType.Atoms
    _auto_module_class = network_modules.hipnn.Hipnn

    @_parent_expander.match(_BaseNode, PairIndexer)
    def expansion2(self, features, pairfinder, **kwargs):
        return features, pairfinder.pair_first, pairfinder.pair_second, pairfinder.pair_dist

    _parent_expander.assertlen(4)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.Atoms, None, None, None)

    def __init__(self, name, parents, periodic=False, module="auto", module_kwargs=None):
        if module == "auto":
            self.module_kwargs = module_kwargs
            net_module = self.auto_module()
        else:
            net_module = module
        parents = self.expand_parents(
            parents, species_set=net_module.species_set, dist_hard_max=net_module.dist_hard_max, periodic=periodic
        )

        super().__init__(name, parents, module=net_module)


class HipnnVec(DefaultNetworkExpansion, AutoKw, Network, SingleNode):
    """
    Node for HIP-NN-TS neural networks, l=1
    """

    _input_names = "input_features", "pair_first", "pair_second", "pair_dist", "pair_coord"
    _index_state = IdxType.Atoms
    _auto_module_class = network_modules.hipnn.HipnnVec

    @_parent_expander.match(_BaseNode, PairIndexer)
    def expansion2(self, features, pairfinder, **kwargs):
        return features, pairfinder.pair_first, pairfinder.pair_second, pairfinder.pair_dist, pairfinder.pair_coord

    _parent_expander.assertlen(5)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.Atoms, None, None, None, None)

    def __init__(self, name, parents, periodic=False, module="auto", module_kwargs=None):
        if module == "auto":
            self.module_kwargs = module_kwargs
            net_module = self.auto_module()
        else:
            net_module = module
        parents = self.expand_parents(
            parents, species_set=net_module.species_set, dist_hard_max=net_module.dist_hard_max, periodic=periodic
        )

        super().__init__(name, parents, module=net_module)


class HipnnQuad(HipnnVec):
    """
    Node for HIP-NN-TS neural networks, l=2
    """

    _auto_module_class = network_modules.hipnn.HipnnQuad



import torch
import threading
from functools import partial
import collections
from typing import List

@torch.jit.script
def get_id_and_starts(key):
    n_items = key.shape[0]
    key_diff = key[:-1] - key[1:]
    key_start = torch.nonzero(key_diff)[:,0] + 1
    start = torch.zeros((1,),device=key.device,dtype=torch.long)
    end = n_items*torch.ones((1,),device=key.device,dtype=torch.long)
    key_start = torch.cat([start,key_start,end])
    key_ids = key[key_start[:-1]]
    return key_ids, key_start

@torch.jit.script
def resort_pairs(key,others: List[torch.Tensor]):
    keysort, argsort = torch.sort(key)
    others = [o[argsort] for o in others]
    key_ids, key_starts = get_id_and_starts(keysort)
    return argsort, key_ids, key_starts, keysort, others

class _CacheEntry():
    def __init__(self,key,cache):
        # Cache key is stored this way because sometimes 
        # pytorch 'resurrects' a Tensor from C++ to pytorch
        # during an autograd calculation as a completely
        # new pyobject, so id(key) is not a completely safe way
        # to know if it's the same tensor.
        # This effect seems to go away when tensor is cached
        # (possibly because the pyobject is not garbage collected)
        # But can we do better to ensure we have the exact same data
        # as in the previous cache?
        # This could pose a problem if, for example,
        # we wrote new pairs in-place during MD and there
        # happened to be the same number with a different structure.
        # Explicitly checking that the key is exactly equal
        # is problematic because it requires device synchronization
        # in order to proceed in the interpreter.
        self.cache_key = key.data_ptr(), key.shape[0], key.device
        self.key = key
        self.cache = cache
        self.computed = False
   
    def __eq__(self,other):
        if not isinstance(other,type(self)):
            return False    
        if self.key is other.key:
            return True
        if self.cache_key == other.cache_key:
            # For now raise an error since this feature is experimental.
            # Most of the benefits seem to work from the `is` check.
            if not torch.equal(self.key,other.key):
                raise RuntimeError("Caching by key does not work! Please raise an issue.")
            return True 
        return False
    
    def find(self):
        if self in self.cache:
            other =  self.cache[self.cache.index(self)]
            return other
        for other in self.cache:
            # Expensive, but not as expensive as sorting + nonzero,
            # which is also a blocking operation.
            # Note that pytorch is smart enough not to generate 
            # a blocking call unless the actual memory needs to be
            # checked, i.e. it will skip if arrays have different sizes.
            if torch.equal(self.key,other.key):
                return other
        else:
            return self    
    
    def compute_and_store(self):
        if self.computed:
            # This shouldn't happen! The guard is here because
            # this code path could activate if someone calls this class in the
            # wrong way from outside the clss.
            raise RuntimeError("A key that was marked computed was asked to be re-computed.")
        keysort, argsort = torch.sort(self.key)
        key_ids, key_starts = get_id_and_starts(keysort)
        self.argsort = argsort
        self.key_ids = key_ids
        self.key_starts = key_starts
        self.computed = True
        self.cache.append(self)
    
    def retrieve(self,others):
        keysort = self.key[self.argsort]
        others = [o[self.argsort] for o in others]
        return self.argsort, self.key_ids, self.key_starts, keysort, others

    @classmethod
    def lookup_key(cls,key,cache):
        entry = cls(key,cache)
        return entry.find()

N_CACHED_KEYS_PER_DEVICE = 2
def _make_cache():
    deque = collections.deque(maxlen=N_CACHED_KEYS_PER_DEVICE)
    lock = threading.Lock()
    return deque, lock
# Dict mapping device to key cache info
_CACHE_STORE = collections.defaultdict(_make_cache)

CACHE_LOCK_MISSES = 0
def resort_pairs_cached(key, others):
    global CACHE_LOCK_MISSES
    deque, lock = _CACHE_STORE[key.device]
    got_lock = lock.acquire(blocking=False)
    if got_lock:
        entry = _CacheEntry.lookup_key(key,deque)
        if not entry.computed:
            entry.compute_and_store()
        lock.release()
        return entry.retrieve(others)
    CACHE_LOCK_MISSES += 1
    return resort_pairs(key,others)




import torch
import numpy as np

# np.set_printoptions(threshold=np.inf)

debug = False
output_dist = False


def check_gradient(x, tag="tag"):
    if torch.isnan(x).any():
        print(x.detach().cpu().data.numpy())
        raise ValueError(tag + ": nan in gradient")
    if torch.isinf(x).any():
        print(x.detach().cpu().data.numpy())
        raise ValueError(tag + ": inf in gradient")


def check(x, tag="tag"):
    if debug:
        if torch.isnan(x).any():
            print(x.detach().cpu().data.numpy())
            raise ValueError(tag + ": nan in tensor")
        if torch.isinf(x).any():
            print(x.detach().cpu().data.numpy())
            raise ValueError(tag + ": inf in tensor")
        if x.requires_grad:
            x.register_hook(lambda grad: check_gradient(grad, tag=tag))
    else:
        pass


def save(x, name="name"):
    np.save(name + ".npy", x.detach().cpu().data.numpy())


def check_dist(x, tag="name"):
    if output_dist:
        with torch.no_grad():
            print(tag + " abs max: " + str(x.abs().max(dim=0)[0]))
            print(tag + " mean   : " + str(x.mean(dim=0)))
            print(tag + " std    : " + str(x.std(dim=0)))
    else:
        pass



"""
Layers for simple operations
"""
import torch


class LambdaModule(torch.nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def extra_repr(self):
        return self.fn.__name__

    def forward(self, *args, **kwargs):
        return self.fn(*args, **kwargs)


class ListMod(torch.nn.Module):
    def forward(self, *features):
        return [x.to(torch.get_default_dtype()) for x in features]


class AtLeast2D(torch.nn.Module):
    def forward(self, item):
        if item.ndimension() == 1:
            item = item.unsqueeze(1)
        return item


class ValueMod(torch.nn.Module):
    def __init__(self, value):
        super().__init__()
        if not isinstance(value, torch.Tensor):
            value = torch.tensor(value, dtype=torch.get_default_dtype())

        self.register_buffer("value", value)

    def extra_repr(self):
        return str(self.value.data)

    def forward(self):
        return self.value


class Idx(torch.nn.Module):
    def __init__(self, index, repr_info=None):
        super().__init__()
        self.index = index
        self.repr_info = repr_info

    def extra_repr(self):
        if self.repr_info is None:
            return ""
        return "{parent_name}.{index}".format(**self.repr_info)

    def forward(self, bundled_inputs):
        return bundled_inputs[self.index]



"""
Nodes for atom index state conversion
"""
from .. import find_unique_relative
from ..indextypes import register_index_transformer, IdxType, _debprint as idx_debprint
from ..nodes.base.node_functions import NodeOperationError
from ..nodes.indexers import PaddingIndexer, OneHotEncoder, AtomReIndexer, AtomDeIndexer
from ..nodes.inputs import SpeciesNode


# TODO: Rewrite so it can use non-one-hot-encodings?


@register_index_transformer(IdxType.MolAtom, IdxType.Atoms)
def idx_molatom_atom(node):
    purpose = "auto-generating indexing for {}".format(node)

    funrel = find_unique_relative  # Abbreviation because we so many calls in this function.
    if node.origin_node is None:
        # If we are auto-indexing in the model graph, it should be easy.
        # species = funrel(node, SpeciesNode,
        #                 why_desc=purpose)
        pidxer = funrel(node, PaddingIndexer, why_desc=purpose)
        idx_debprint("Using reindexer for ", node)

    else:
        # If we are auto-indexing in the loss graph, it gets a bit complicated.

        # The species we link to will be the true version of species, that is, the node where
        species = funrel(node.origin_node, SpeciesNode, why_desc=purpose).true
        try:
            encoder = funrel(species, OneHotEncoder, why_desc=purpose)
        except NodeOperationError as ne:
            idx_debprint("Creating new ENCODER")
            # If this fails, something bad has happened -- the loss graph is trying to do something not defined by the
            # model graph
            origin_encoder = funrel(species.origin_node, OneHotEncoder, why_desc=purpose)
            encoder = OneHotEncoder("Auto(One-hot)", (species,), species_set=origin_encoder.species_set)
            # If we can't find an encoder, let's assume we won't find an indexer
            pidxer = PaddingIndexer("Auto(PaddingIndexer)", encoder)
        else:
            # If we did find an encoder, look for an indexer and assume it exists
            pidxer = funrel(encoder, PaddingIndexer, why_desc=purpose)

    cls = AtomReIndexer
    parents = node, pidxer.real_atoms
    return parents, cls


@register_index_transformer(IdxType.Atoms, IdxType.MolAtom)
def idx_atom_molatom(node):

    if node.origin_node is None:
        parents = (node,)
        return parents, AtomDeIndexer
    else:
        raise NotImplementedError("De-indexing not yet implemented in loss graph")
        # TODO: refactor out padding indexer creation for loss from the molatom-atom indexer, then re-use it here.



"""
Layers for target prediction such as Energy and Charge
"""
import torch

from . import indexers
from . import hiplayers


class HEnergy(torch.nn.Module):
    """
    Predict a system-level scalar such as energy from a sum over local components.
    """

    def __init__(self, feature_sizes, first_is_interacting=False, n_target=1):

        super().__init__()
        self.first_is_interacting = first_is_interacting
        if first_is_interacting:
            feature_sizes = feature_sizes[1:]

        self.feature_sizes = feature_sizes

        self.summer = indexers.MolSummer()
        self.n_terms = len(feature_sizes)
        biases = (first_is_interacting, *(True for _ in range(self.n_terms - 1)))

        self.layers = torch.nn.ModuleList(
            torch.nn.Linear(nf, n_target, bias=bias) for nf, bias in zip(feature_sizes, biases)
        )

    def forward(self, all_features, mol_index, n_molecules):
        """
        Pytorch Enforced Forward function

        :param: all_features a list of feature tensors:
        :param: mol_index the molecular index for atoms in the batch
        :param: total number of molecules in the batch
        :return: Total Energy
        """
        if self.first_is_interacting:
            all_features = all_features[1:]

        partial_energies = [lay(x) for x, lay in zip(all_features, self.layers)]
        partial_terms = [self.summer(x, mol_index, n_molecules) for x in partial_energies]
        partial_sums = [partial_terms[0]]
        z = partial_terms[0]
        for x in partial_terms[1:]:
            z = x + z
            partial_sums.append(z)

        total_atomen = sum(partial_energies)
        total_energies = self.summer(total_atomen, mol_index, n_molecules)

        if self.n_terms > 1:
            partial_esq = [torch.square(x) for x in partial_energies]
            partial_atom_hier = [x / (x + y) for x, y in zip(partial_esq[1:], partial_esq[:-1])]
            mol_hier = [self.summer(x, mol_index, n_molecules)/self.summer(x+y, mol_index, n_molecules)
                        for x,y in zip(partial_esq[1:], partial_esq[:-1])]
            mol_hier = sum(mol_hier)
            partial_batch_hier = [x.sum() / (x.sum() + y.sum()) for x, y in zip(partial_esq[1:], partial_esq[:-1])]
            batch_hier = sum(partial_batch_hier)
            total_atom_hier = sum(partial_atom_hier)
            total_hier = self.summer(total_atom_hier, mol_index, n_molecules)

        else:
            total_hier = torch.zeros_like(total_energies)
            mol_hier = torch.zeros_like(total_energies)
            total_atom_hier = torch.zeros_like(total_atomen)
            batch_hier = torch.zeros(1,dtype=total_energies.dtype,device=total_energies.device)

        return total_energies, total_atomen, partial_sums, total_hier, total_atom_hier, mol_hier, batch_hier


class HCharge(torch.nn.Module):
    """
    Predict an atom-level scalar such as charge from local features.
    """

    def __init__(self, feature_sizes, first_is_interacting=False, n_target=1):
        super().__init__()
        self.feature_sizes = feature_sizes

        self.n_terms = len(feature_sizes)
        self.n_target = n_target
        biases = (first_is_interacting, *(True for _ in range(self.n_terms - 1)))
        self.layers = torch.nn.ModuleList(
            torch.nn.Linear(nf, n_target, bias=bias) for nf, bias in zip(feature_sizes, biases)
        )

    def forward(self, all_features):
        """
        :param all_features a list of feature tensors:
        :return: charges, predicted charges summed at each layer, and charge hierarchcality
        """
        partial_charges = [lay(x) for x, lay in zip(all_features, self.layers)]
        partial_sums = [partial_charges[0]]
        z = partial_charges[0]
        for x in partial_charges[1:]:
            z = x + z
            partial_sums.append(z)

        total_charges = sum(partial_charges)

        if self.n_terms > 1:
            partial_esq = [x ** 2 for x in partial_charges]
            partial_atom_hier = [x / (x + y) for x, y in zip(partial_esq[1:], partial_esq[:-1])]
            charge_hier = sum(x for x in partial_atom_hier)
        else:
            charge_hier = None

        return total_charges, partial_sums, charge_hier


class LocalChargeEnergy(torch.nn.Module):
    def __init__(self, feature_sizes, first_is_interacting=False):
        super().__init__()
        self.feature_sizes = feature_sizes

        self.n_terms = len(feature_sizes)
        bias_state = (first_is_interacting, *(True for _ in range(self.n_terms - 1)))
        self.layers_lin = torch.nn.ModuleList(
            torch.nn.Linear(nf, 1, bias=bias) for nf, bias in zip(feature_sizes, bias_state)
        )
        self.layers_quad = torch.nn.ModuleList(
            torch.nn.Linear(nf, 1, bias=bias) for nf, bias in zip(feature_sizes, bias_state)
        )
        self.summer = indexers.MolSummer()

    def forward(self, charges, all_features, mol_index, n_molecules):

        partial_lin_terms = [lay(x) for x, lay in zip(all_features, self.layers_lin)]
        partial_quad_terms = [lay(x) for x, lay in zip(all_features, self.layers_lin)]

        total_lin = sum(partial_lin_terms)
        total_quad = sum(partial_quad_terms)

        atom_charge_energy = (total_quad * charges) ** 2 + total_lin
        molecule_charge_energy = self.summer(atom_charge_energy, mol_index, n_molecules)

        return molecule_charge_energy, atom_charge_energy


class HBondSymmetric(torch.nn.Module):
    def __init__(
        self,
        feature_sizes,
        n_dist,
        dist_soft_min,
        dist_soft_max,
        dist_hard_max,
        positive=False,
        symmetric=False,
        antisymmetric=False,
        sensitivity_type=hiplayers.InverseSensitivityModule,
        n_target=1,
        all_pairs=True,
    ):
        super().__init__()

        if symmetric and antisymmetric:
            raise ValueError("Bond-like prediction cannot be both symmetric and antisymmetric!")

        if antisymmetric and positive:
            raise ValueError("Bond-like prediction cannot be antisymmetric and positive!")

        self.sensitivity = sensitivity_type(n_dist, dist_soft_min, dist_soft_max, dist_hard_max)
        self.feature_sizes = feature_sizes
        self.n_terms = len(feature_sizes)

        self.symmetric = symmetric
        self.antisymmetric = antisymmetric
        self.positive = positive
        self.n_target = n_target
        self.n_dist = n_dist
        if positive:
            self.biases = torch.nn.ParameterList(torch.nn.Parameter(torch.zeros(n_target)) for _ in feature_sizes)
        else:
            self.biases = None

        self.weights = torch.nn.ParameterList(
            torch.nn.Parameter(torch.zeros(n_dist, n_target, nf, nf)) for nf in feature_sizes
        )
        for p in self.weights:
            torch.nn.init.xavier_normal_(p.data)

    def forward(self, all_features, pair_first, pair_second, pair_dist):

        weights = self.weights
        if self.symmetric:
            weights = [w + w.transpose(2, 3) for w in weights]
        if self.antisymmetric:
            weights = [w - w.transpose(2, 3) for w in weights]

        sense_vals = self.sensitivity(pair_dist)

        n_d, n_t, n_f, _ = self.weights[0].shape
        # These are the contributions for each bond at a given sensitivity distance
        # bilinear takes shape (pair,feature1),(pair,feature2),(ndist*n_target,feature1,feature2)
        # and sums to pair,(ndist*n_target), which is reshaped.
        partial_bond_dists = [
            torch.nn.functional.bilinear(
                f[pair_first],
                f[pair_second],
                w.reshape(self.n_dist * self.n_target, f.shape[-1], f.shape[-1]),
                bias=None,
            ).reshape(-1, self.n_dist, self.n_target)
            for f, w in zip(all_features, weights)
        ]

        # These are the contributions for each bond
        # multiply pair,ndist by pair,ndist,n_targets
        partial_bonds = [(pbd * sense_vals.unsqueeze(2)).sum(dim=1) for pbd in partial_bond_dists]
        # NOTE:
        # Einsum implementation of combined partial_bond_dists and partial_bond operations, seems to be slower.
        # But leaving this comment here for readability.
        # partial_bonds = [
        #     torch.einsum("bf,bg,bs,stfg->bt",f[pair_first],f[pair_second],sense_vals,w)
        #     for f,w in zip(all_features,weights)
        # ]

        if self.positive:
            partial_bonds = [pb + b for pb, b in zip(partial_bonds, self.biases)]

        total_bonds = sum(partial_bonds)
        if self.positive:
            total_bonds = torch.nn.functional.softplus(total_bonds)

        if self.n_terms > 1:
            partial_bsq = [b ** 2 for b in partial_bonds]
            partial_hier = [x / (x + y) for x, y in zip(partial_bsq[1:], partial_bsq[:-1])]
            bond_hier = sum(partial_hier)
        else:
            bond_hier = torch.zeros_like(total_bonds)
        return total_bonds, bond_hier





import setuptools
import versioneer

with open("README.rst", "r") as fh:
    long_description = fh.read()

doc_requirements = [
    "sphinx",
    "sphinx_rtd_theme",
    "ase",
]

full_requirements = [
    "ase",
    "numba",
    "matplotlib",
    "tqdm",
    "graphviz",
    "h5py",
]

setuptools.setup(
    name="hippynn",
    version=versioneer.get_version(),
    author="Nicholas Lubbers et al",
    author_email="hippynn@lanl.gov",
    python_requires=">=3.9",
    install_requires=[
        "numpy",
        "torch",
    ],
    extras_require={"docs": doc_requirements, "full": full_requirements},
    license="BSD 3-Clause License",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries",
    ],
    description="The hippynn python package - a modular library for atomistic machine learning with pytorch",
    long_description=long_description,
    packages=setuptools.find_packages(),
    cmdclass=versioneer.get_cmdclass(),
)
